/*
    Copyright (C) 1999, 2000, 2001, 2002, 2003  Charles MacDonald

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "shared.h"

extern int32 genZ_specialfskip;

t_bitmap bitmap;
t_input input;
t_snd snd;
//static int sound_tbl[262];


extern t_YM2612_logW YM2612_logW[0x10000]; //not more than 0x1000 write/sound update
extern uint32 YM2612_log_cpt;
extern t_SN76496_logW SN76496_logW[0x10000]; //not more than 0x1000 write/sound update
extern uint32 SN76496_log_cpt;

#define SND_SIZE (snd.buffer_size * sizeof(int16))

void audio_update0(void);

int audio_init(int rate,int buffersize)
{
    int i;

    /* 68000 and YM2612 clock */
    float vclk = 53693175.0 / 7;

    /* Z80 and SN76489 clock */
    float zclk = 3579545.0;

    /* Clear the sound data context */
    memset(&snd, 0, sizeof(snd));

    /* Make sure the requested sample rate is valid */
    if(!rate || ((rate < 8000) | (rate > 44100)))
    {
        return (0);
    }

    /* Calculate the sound buffer size */
    snd.buffer_size = buffersize;
    snd.sample_rate = rate;

    /* Allocate sound buffers */
	
    

    snd.fm.buffer[0] = ljz_malloc(snd.buffer_size * sizeof(int16));
    snd.fm.buffer[1] = ljz_malloc(snd.buffer_size * sizeof(int16));        
    snd.psg.buffer = ljz_malloc(snd.buffer_size * sizeof(int16));
    
    

    /* Make sure we could allocate everything */
    if(/*!snd.buffer[0] || !snd.buffer[1] ||*/ !snd.fm.buffer[0] || !snd.fm.buffer[1] || !snd.psg.buffer)
    {
        return (0);
    }

    /* Initialize sound chip emulation */
    SN76496_sh_start(zclk, 100, rate);
    YM2612Init(1, vclk, rate, NULL, NULL);

    /* Set audio enable flag */
    snd.enabled = 1;
    snd.fastmode=0;

    /* Make sound table */
/*    for (i = 0; i < 262; i++)
    {
        float p = snd.buffer_size * i; //todo adapt for nrsc/pal 60/50
        p = p / 262;
    	sound_tbl[i] = p;
    }    	
  */  
//    snd.fm.lastStage = snd.fm.curStage = 0;
//    snd.psg.lastStage = snd.psg.curStage = 0;
    
    YM2612_log_cpt=0;
	SN76496_log_cpt=0;
	snd.total_cycles=0;

    return (0);    
}

void audio_fastmode(int fast)
{
	if (snd.enabled)
	{
		snd.fastmode=fast;
		YM2612_log_cpt=0;
		SN76496_log_cpt=0;
		snd.total_cycles=0;
	}
}

void system_init(void)
{
    gen_init();
    vdp_init();
    render_init();
}

void system_reset(void)
{
    gen_reset();
    vdp_reset();
    render_reset();

    if(snd.enabled)
    {
        YM2612ResetChip(0);
//        memset(snd.buffer[0], 0, SND_SIZE);
//        memset(snd.buffer[1], 0, SND_SIZE);
        memset(snd.fm.buffer[0], 0, SND_SIZE);
        memset(snd.fm.buffer[1], 0, SND_SIZE);
        memset(snd.psg.buffer, 0, SND_SIZE);
        
//        memset(snd.fm.bufferB[0], 0, SND_SIZE);
//        memset(snd.fm.bufferB[1], 0, SND_SIZE);
//        memset(snd.psg.bufferB, 0, SND_SIZE);
    }
}

void audio_shutdown(void)
{
	if (snd.enabled)
    {
	    ljz_free(snd.fm.buffer[0]);
	    ljz_free(snd.fm.buffer[1]);    
    	ljz_free(snd.psg.buffer);        	
    	YM2612Shutdown();
    }
}

void system_shutdown(void)
{
    gen_shutdown();
    vdp_shutdown();
    render_shutdown();  
    if (snd.enabled) audio_shutdown();
}


int system_frame(int do_skip)
{
    int line;
    int total=0,aim=0,executed;

    if(gen_running == 0)
        return 0;

    /* Clear V-Blank flag */
    status &= ~0x0008;

    /* Toggle even/odd flag (IM2 only) */
    if(im2_flag)
        status ^= 0x0010;

    /* Point to start of sound buffer */
/*    snd.fm.lastStage = snd.fm.curStage = 0;
    snd.psg.lastStage = snd.psg.curStage = 0;*/

    /* Parse sprites for line 0 (done on line 261) */
    parse_satb(0x80);
        

    for(line = 0; line < 262; line += 1)
    {
        /* Used by HV counter */
        v_counter = line;

        /* Run 68000 emulation */
        
		if (genZ_cpu_cyclone)
		{
        aim+=487;
        executed=CycloneDoRun(aim-total);
        total_cycles+=executed;
        snd.total_cycles+=executed;
        total+=executed;
        }
		else 
		{
			m68k_execute(487);
			snd.total_cycles+=487;
		}

        
        
        if(!gen_running) break;

        /* Run Z80 emulation (if enabled) */
        if(zreset == 1 && zbusreq == 0)
        {        	
            if (emul_z80) z80_execute(228);
            if(!gen_running) break;
        }

        /* If a Z80 interrupt is still pending after a scanline, cancel it */
        if(zirq == 1)
        {
            zirq = 0;
            if (emul_z80) z80_set_irq_line(0, CLEAR_LINE);
        }

        /* Render a line of the display */
        if(do_skip == 0)
        {
        	if (genZ_specialfskip)
        	{
        		if (genZ_framecount&1)
        		{
	        		if ( (line<frame_end)&&(line>=(frame_end>>1)) ) render_line(line);
					if ( (line<frame_end-1)&&(line>=(frame_end>>1)-1) ) parse_satb(0x81 + line);
        		}
        		else
        		{
	        		if(line < (frame_end>>1) ) render_line(line);
	        		if(line < (frame_end>>1) ) parse_satb(0x81 + line);
        		}
        	}
        	else
        	{
	            if(line <  frame_end   ) render_line(line);
	            if(line <  frame_end-1 ) parse_satb(0x81 + line);
	        }       		
	        
    	    
    	    
        }

        /* Do H interrupt processing */
        if(line <= frame_end)
        {
            counter -= 1;
            if(counter == -1)
            {
                counter = reg[10];
                hint_pending = 1;
                if(reg[0] & 0x10)
                {
					if (genZ_cpu_cyclone)
                    CycloneInterrupt(4);
					else
					m68k_set_irq(4);

                }
            }
        }
        else
        {
            counter = reg[10];
        }

        /* Do end of frame processing */
        if(line == frame_end)
        {
            status |= 0x0088;
            vint_pending = 1;

            /* Give enough time to read the interrupt pending flag before
               the interrupt actually occurs. */
			if (genZ_cpu_cyclone)
			{
            aim+=16;
	        executed=CycloneDoRun(aim-total);
    	    total_cycles+=executed;    	    
    	    snd.total_cycles+=executed;
        	total+=executed;            
        	}
			else
			{            
            	m68k_execute(16);
            	snd.total_cycles+=16;            	
            }

        	
            if(!gen_running) break;

            if(reg[1] & 0x20)
            {
			if (genZ_cpu_cyclone)
                CycloneInterrupt(6);
			else
				m68k_set_irq(6);

            }

            if(zreset == 1 && zbusreq == 0)
            {
                if (emul_z80) z80_set_irq_line(0, ASSERT_LINE);
                zirq = 1;
            }
        }
 
        //fm_update_timers();
        YM2612_logW[YM2612_log_cpt].cycles=snd.total_cycles;
        YM2612_logW[YM2612_log_cpt].address=4;
        YM2612_logW[YM2612_log_cpt++].data=0;
        YM2612_log_cpt&=0xFFFF;

/*		snd.fm.curStage = sound_tbl[line];
        snd.psg.curStage = sound_tbl[line];*/
    }

/*    if(snd.enabled)
    {
        audio_update0();
    }*/

    return gen_running;
}






void audio_update(int16 *buffer)
{
	int i,j;
    int16 acc;
    int16 *tempBuffer[2];
    int16 psg;
    
    int32 event_cpt;
    int32 cur_cycle,next_cycle;
    int32 bsize,tsize;
    
    if (snd.fastmode)
    {
		tempBuffer[0] = snd.fm.buffer[0];
	    tempBuffer[1] = snd.fm.buffer[1];
	    YM2612UpdateOne(0, (int16 **)tempBuffer, snd.buffer_size);
		tempBuffer[0] = snd.psg.buffer;
		SN76496Update(0, tempBuffer[0], snd.buffer_size);
    }
	else
	{	
	if (!snd.total_cycles)
	{
		for(i = 0; i < snd.buffer_size; i += 1)
    	{       	
	        *buffer++=0;	    
	        *buffer++=0;
	    }	   
	    return;
	}	
	
    tsize=0;
    if (YM2612_log_cpt)
    {   
	    cur_cycle=0; 
    	for (event_cpt=0;event_cpt<YM2612_log_cpt;event_cpt++)
    	{    	
    		next_cycle=YM2612_logW[event_cpt].cycles;    		
    		if (YM2612_logW[event_cpt].address==4) fm_update_timers();
    		else
    		{    										
			    bsize=snd.buffer_size*(next_cycle-cur_cycle)/snd.total_cycles;
				if (bsize>4)
				{
				    if ((tsize+bsize)<=snd.buffer_size)
				    {
				    	tempBuffer[0] = snd.fm.buffer[0]+tsize;
					    tempBuffer[1] = snd.fm.buffer[1]+tsize;
				    	YM2612UpdateOne(0, (int16 **)tempBuffer, bsize);
	   				    tsize+=bsize;
   					    cur_cycle=next_cycle;
				    }
				    else
				    if (tsize<snd.buffer_size)
				    {			    	
				    	tempBuffer[0] = snd.fm.buffer[0]+tsize;
					    tempBuffer[1] = snd.fm.buffer[1]+tsize;
     				    YM2612UpdateOne(0, (int16 **)tempBuffer, snd.buffer_size-tsize);
						tsize=snd.buffer_size;
						cur_cycle=next_cycle;
				    }
			    }
			    YM2612Write(0, YM2612_logW[event_cpt].address, YM2612_logW[event_cpt].data);
			}
			
		}
		
		if (tsize<snd.buffer_size)
		{
			tempBuffer[0] = snd.fm.buffer[0]+tsize;
		    tempBuffer[1] = snd.fm.buffer[1]+tsize;
		    YM2612UpdateOne(0, (int16 **)tempBuffer, snd.buffer_size-tsize);
		}
		YM2612_log_cpt=0;
	}
    
    
    tsize=0;
    if (SN76496_log_cpt)
    {   
	    cur_cycle=0; 
    	for (event_cpt=0;event_cpt<SN76496_log_cpt;event_cpt++)
    	{    	
    		next_cycle=SN76496_logW[event_cpt].cycles;    					
		    bsize=snd.buffer_size*(next_cycle-cur_cycle)/snd.total_cycles;
			if (bsize>4)
			{
			    if ((tsize+bsize)<=snd.buffer_size)
			    {
			    	tempBuffer[0] = snd.psg.buffer+tsize;
			    	SN76496Update(0, tempBuffer[0], bsize);
   				    tsize+=bsize;
   				    cur_cycle=next_cycle;
			    }
			    else
			    if (tsize<snd.buffer_size)
			    {			    	
				    tempBuffer[0] = snd.psg.buffer+tsize;
     			    SN76496Update(0, tempBuffer[0], snd.buffer_size-tsize);
					tsize=snd.buffer_size;
					cur_cycle=next_cycle;
			    }
		    }
		    SN76496Write(0, SN76496_logW[event_cpt].data);					
		}
		
		if (tsize<snd.buffer_size)
		{
			tempBuffer[0] = snd.psg.buffer+tsize;
		    SN76496Update(0, tempBuffer[0], snd.buffer_size-tsize);
		}
		SN76496_log_cpt=0;
	}
	snd.total_cycles=0;
	}
	
    for(i = 0; i < snd.buffer_size; i += 1)
    	{
       		psg = snd.psg.buffer[i]/2;
	        acc = snd.fm.buffer[0][i];
    	    acc += psg;        
	        *buffer++=acc>>1;
	        acc = snd.fm.buffer[1][i];
	        acc += psg;    	
	        *buffer++=acc>>1;
	    }	   
		
}
