
#include "shared.h"
#include "LJZmenu.h"


#define KEY_UP    		0x00000001
#define KEY_DOWN  		0x00000002
#define KEY_LEFT  		0x00000004
#define KEY_RIGHT 		0x00000008
#define KEY_SEL   		0x00000010
#define KEY_A     		0x00000020
#define KEY_B	  		0x00000040
#define KEY_C	 		0x00000080
#define KEY_START		0x00000100
#define KEY_QUICKLOAD   0x00000200
#define KEY_QUICKSAVE   0x00000400
#define KEY_TURBO  		0x00000800
#define KEY_SCREENMODE	0x00001000
#define KEY_FRAMESKIP	0x00002000




// Assertion Support
static void
assert_check(int error, const char* message)
{
    if (error) {
        ErrDisplayFileLineMsg("PalmOS", (UInt16) error, message);
    }
}

#define xassert(error) {assert_check(error, #error);}

static void showError(const char* aWhat, Err aError) {
    ErrDisplayFileLineMsg("oink", (UInt16) aError, aWhat);
}


char str_month[12][4]=
{"Jan","Feb","Mar","Apr","May","Jun",
"Jul","Aug","Sep","Oct","Nov","Dec"};

char RomPath[260];
char shortrom_filename[64];
char genZ_curVolume;

int genZ_loadrom();
void genZ_end();
int genZ_setupenv(TwGfxType* mGfxLib);

void genZ_changefskip(void);
void genZ_changerendering(void);
void genZ_updatevpad(unsigned int val);
void genZ_MainLoop(void);
void genZ_pause(int pause);
int genZ_ShowMenu(void);
void genZ_updateScreen(void);
void genZ_asmfastrender256(uint8 *screen,uint8 *source,uint32 h,uint32 ofs);
void genZ_asmfastrender320(uint8 *screen,uint8 *source,uint32 h,uint32 ofs);
int32 genZ_getTickCount(void);

TwGfxType *genZ_mGfxLib;
TwGfxSurfaceType *genZ_curRenderingScreen,*genZ_RenderingScreen[2];
uint16 *genZ_curRenderingScreenPtr,*genZ_RenderingScreenPtr[2];
uint32 genZ_curRenderingScreenIndex;
uint32 genZ_enabledSound,genZ_sndfreq;
uint32 genZ_renderingmode;
uint32 genZ_padState;
int32 genZ_frameskipped,genZ_frameskip,genZ_fastforward,genZ_specialfskip;
int32 genZ_frameBeginTime;
int32 genZ_speedpct;
int32 genZ_autofireA,genZ_autofireB,genZ_autofireC;
int32 genZ_framecount,genZ_drawframe;

int genZ_cpu_cyclone;

int32 genZ_frameTime;

TwInputHandle genZ_mInputQueue;
Int32 genZ_inputFormats[] = { twInputTimeStamp };
typedef struct {Int32 mTime;} genZ_ZUserEvent;
genZ_ZUserEvent genZ_mUserEvent;
Int32* genZ_mInputFormat;        

int state_load(int num,int onlytest);
int state_save(int num);

uint8 VideoBuffer[512*512*2];


static unsigned char *RomData;
static int RomLen;

#include "SoundMgr.h"
static SndStreamRef snd_stream;
typedef struct {uint32 __r9;uint32 __r10;} reg_sys_t;
static reg_sys_t reg_sys;
uint32 genZ_sndbuffsize;

#include "sys/stat.h"


static Err genZ_sndCallBack (void *userData,SndStreamRef stream, void *bufferZ,UInt32 frameCountZ)
{
	__asm 
	{
		stmfd  r13!,{r9,r10}
		ldr	   r9,[r0]
		ldr	   r10,[r0,#4]
	}		
	//fm_update_timers();
	audio_update((int16*)bufferZ);
	__asm 
	{
		ldmfd  r13!,{r9,r10}
	}
	return errNone;
}

int update_input(void);

void load_sram(void)
{
	FILE *f;	
	char savename[255],result[256];
	uint8 i=0,j=255,k=255,l=255;
	char chaine[256];  
	i=0;
	j=255;
	k=-1;
	while (RomPath[i]!=0) 
	{
		if (RomPath[i]=='.') j=i;
		if (RomPath[i]=='/') k=i;
		if (RomPath[i]=='\\') k=i;
		i++;
	  }
	if (j==255) j=i;  //pas d'extension	  
	memcpy(result,RomPath+k+1,j+1-k-1);    
	result[j+1-k-1]=0;
	strcat(result,"srm");
	sprintf(savename,"%s%s",GENZ_SAVESDIR,result);
	savename[0]=genZ_curVolume+'0';
	
	f=fopen(savename,"rb");
	if (!f) return;
	fread(saveram,save_len,1,f);
	fclose(f);
}
void save_sram(void)
{
	FILE *f;	
	char savename[255],result[256];
	uint8 i=0,j=255,k=255,l=255;
	char chaine[256];  
	i=0;
	j=255;
	k=-1;
	while (RomPath[i]!=0) 
	{
		if (RomPath[i]=='.') j=i;
		if (RomPath[i]=='/') k=i;
		if (RomPath[i]=='\\') k=i;
		i++;
	  }
	if (j==255) j=i;  //pas d'extension	  
	memcpy(result,RomPath+k+1,j+1-k-1);    
	result[j+1-k-1]=0;
	strcat(result,"srm");
	sprintf(savename,"%s%s",GENZ_SAVESDIR,result);
	savename[0]=genZ_curVolume+'0';
	f=fopen(savename,"w+b");
	if (!f) return;
	fwrite(saveram,save_len,1,f);
	fclose(f);
}

int update_input(void)
{
    int running = 1;
    
  int pad=0;
  
  if (genZ_padState&KEY_QUICKLOAD)
	{
		if (!state_load(0,1))
		{
			if (menu_confirm("Load state ?")) state_load(0,0);
			genZ_frameBeginTime=genZ_getTickCount()+genZ_frameTime*100/genZ_speedpct;
		}
	}
	if (genZ_padState&KEY_QUICKSAVE)
	{
		if (menu_confirm("Save state ?")) state_save(0);
			genZ_frameBeginTime=genZ_getTickCount()+genZ_frameTime*100/genZ_speedpct;
	}
	if (genZ_padState&KEY_TURBO)
	{		
		genZ_fastforward=1;		
	}
	else 
	if (genZ_fastforward)
	{
		genZ_fastforward=0;
		genZ_frameBeginTime=genZ_getTickCount()+genZ_frameTime*100/genZ_speedpct;
	}
    
  if(genZ_padState&KEY_UP)     pad|=INPUT_UP;
  if(genZ_padState&KEY_DOWN)   pad|=INPUT_DOWN;
  if(genZ_padState&KEY_LEFT)   pad|=INPUT_LEFT;
  if(genZ_padState&KEY_RIGHT)  pad|=INPUT_RIGHT;
  if(genZ_padState&KEY_A) 
  {
	 switch (genZ_autofireA)
	 {
  		case 0:pad|=INPUT_A; break;
  		case 1:
  	  		if ((genZ_framecount&31)>15) pad|=INPUT_A;
  	  		break;
  		case 2:
  	  		if ((genZ_framecount&15)>7) pad|=INPUT_A;
  	  		break;
  		case 3:
  	  		if ((genZ_framecount&7)>3) pad|=INPUT_A;
  	  		break;  		   
  	 }
  }  		    
  if(genZ_padState&KEY_B)
  {
 	 switch (genZ_autofireB)
	 {
  		case 0:pad|=INPUT_B; break;
  		case 1:
  	  		if ((genZ_framecount&31)>15) pad|=INPUT_B;
  	  		break;
  		case 2:
  	  		if ((genZ_framecount&15)>7) pad|=INPUT_B;
  	  		break;
  		case 3:
  	  		if ((genZ_framecount&7)>3) pad|=INPUT_B;
  	  		break;  		   
  	 }
  }
  if(genZ_padState&KEY_C)
  {
 	 switch (genZ_autofireC)
	 {
  		case 0:pad|=INPUT_C; break;
  		case 1:
  	  		if ((genZ_framecount&31)>15) pad|=INPUT_C;
  	  		break;
  		case 2:
  	  		if ((genZ_framecount&15)>7) pad|=INPUT_C;
  	  		break;
  		case 3:
  	  		if ((genZ_framecount&7)>3) pad|=INPUT_C;
  	  		break;  		   
  	 }
  }
  if(genZ_padState&KEY_START)  pad|=INPUT_START;

  input.pad[0]=pad;
/*    if(keystate[SDLK_z])      input.pad[0] |= INPUT_X;
    if(keystate[SDLK_x])      input.pad[0] |= INPUT_Y;
    if(keystate[SDLK_c])      input.pad[0] |= INPUT_Z;
    if(keystate[SDLK_v])      input.pad[0] |= INPUT_MODE;*/
 
   //system_reset();

    //if(keystate[SDLK_ESCAPE]) running = 0;
    return (running);
}



static int Do_Nothing(void)
{
	return 0;
}


/*******************************************************/

int32 genZ_getTickCount()
{
	//TODO return tick 
	/*long long l;
	l=TimGetTicks();	
	return l*1000/SysTicksPerSecond();*/	
  	TwInputPoll(genZ_mInputQueue, (Int32 *) &genZ_mUserEvent, sizeof(genZ_mUserEvent));
	return genZ_mUserEvent.mTime;
}


int init_video(void)
{
	//Init vid
	TwGfxSurfaceInfoType surfInfo;
	surfInfo.size = sizeof(surfInfo);
	surfInfo.width = 320; //must be / 4
	surfInfo.height = 240;
	surfInfo.location = twGfxLocationAcceleratorMemoryNoBackingStore;
	surfInfo.pixelFormat = twGfxPixelFormatRGB565;
	xassert(TwGfxAllocSurface(genZ_mGfxLib, &genZ_RenderingScreen[0], &surfInfo));
	xassert(TwGfxAllocSurface(genZ_mGfxLib, &genZ_RenderingScreen[1], &surfInfo));

	if (TwGfxLockSurface(genZ_RenderingScreen[0], (void**)&genZ_RenderingScreenPtr[0]))
	{
	   return 1;
	}
	if (TwGfxLockSurface(genZ_RenderingScreen[1], (void**)&genZ_RenderingScreenPtr[1]))
	{
	   return 1;
	}
	
	memset(genZ_RenderingScreenPtr[0],0,320*240*2);
	memset(genZ_RenderingScreenPtr[1],0,320*240*2);
	
	genZ_curRenderingScreenIndex=0;
	genZ_curRenderingScreen=genZ_RenderingScreen[genZ_curRenderingScreenIndex];
	genZ_curRenderingScreenPtr=genZ_RenderingScreenPtr[genZ_curRenderingScreenIndex];
	
	
	signed long val1,val2,val;
	FtrGet(APPLI_ID,FTR_OPT_SCREENMODE_ID,(unsigned long*)&val1);
	FtrGet(APPLI_ID,FTR_OPT_SMOOTHING_ID,(unsigned long*)&val2);
	if (!val1) genZ_renderingmode=0;
	else 
	if (val1==1) genZ_renderingmode=(val2?2:1);
	else
	if (val1==2) genZ_renderingmode=(val2?4:3);
	else
	genZ_renderingmode=(val2?6:5);
	

	

	genZ_fastforward=0;
	
	genZ_frameskipped=0;
	genZ_specialfskip=0;
	
	 FtrGet(APPLI_ID,FTR_OPT_FSKIP_ID,(unsigned long*)&val); 
	genZ_frameskip=val;
	if (genZ_frameskip==SKEEZIX_FSKIP) genZ_specialfskip=1;
	

    /*input*/
    TwInputOpen(&genZ_mInputQueue, "twinput", "r");
    genZ_mInputFormat = (Int32*)ljz_malloc(sizeof(Int32)*sizeof(genZ_inputFormats)/sizeof(genZ_inputFormats[0]));
    memmove(genZ_mInputFormat, genZ_inputFormats, sizeof(genZ_inputFormats));
    TwInputSetFormat(genZ_mInputQueue, genZ_mInputFormat, sizeof(genZ_inputFormats));    
    TwInputActivate(genZ_mInputQueue);

	return 0;
}


int genZ_setupenv(TwGfxType* mGfxLib)
{
	genZ_mGfxLib=mGfxLib;
	return 0;
}

int genZ_loadrom()
{
	ljz_std_showstat();
	int val;
				
	FtrGet(APPLI_ID,FTR_OPT_SOUND_ID,(unsigned long*)&val);  
	genZ_enabledSound=val;
	FtrGet(APPLI_ID,FTR_OPT_CPUCORE_ID,(unsigned long*)&val);  
	genZ_cpu_cyclone=val;	
	
	FtrGet(APPLI_ID,FTR_OPT_SOUNDFREQ_ID,(unsigned long*)&val);  
	if (val==1) genZ_sndfreq=22050;
	else if (val==2) genZ_sndfreq=11025;
	else genZ_sndfreq=44100;
	
	emul_z80=(genZ_enabledSound&4);
	
	genZ_speedpct=100;
	genZ_autofireA=0;
	genZ_autofireB=0;
	genZ_autofireC=0;
	
	genZ_framecount=0;
	genZ_drawframe=1;
	



	init_video();
	
    char *ftrptr;
    strcpy(RomPath,GENZ_ROMSDIR);    
    FtrGet(APPLI_ID,FTR_ROM_FILENAME_ID ,(unsigned long*)&ftrptr);
    
    genZ_curVolume=ftrptr[0];
    RomPath[0]=genZ_curVolume+'0';
    
    ftrptr+=3;
    strcpy(shortrom_filename,ftrptr);  
    strcat(RomPath,ftrptr);
    FtrPtrFree(APPLI_ID, FTR_ROM_FILENAME_ID);	
	
	// Init emulation:		
    error_init();

	/* Load game */
	char strload[128];
    sprintf(strload,"Loading : %s",shortrom_filename);
    menu_message(strload);

//		unsigned long freeM,maxM;	
//	MemHeapFreeBytes(0,&freeM,&maxM);
//	StrPrintF(strload,"mem : %d %d",freeM,maxM);
//	menu_message(strload);
//	menu_waitPen(NULL,NULL,1);


    if(!load_rom(RomPath))
	{
		char caption[256];
		sprintf(caption, "Error loading file `%s'.",RomPath);
//		MessageBox(NULL, caption, "Error", 0);
		menu_message(caption);
		menu_waitPen(NULL,NULL,1);
//		exit(1);
		return 1;
	}
	
	if (saveram)
	{
		//char str[256];		
		//sprintf(str,"found sram:sz=%d,st=%08X",save_len,save_start);
		//menu_message(str);
		//menu_waitPen(NULL,NULL,1);
		load_sram();
	}

    memset(&bitmap, 0, sizeof(t_bitmap));
    bitmap.width  = 512;
    bitmap.height = 512;
    bitmap.depth  = 16;
    bitmap.granularity = 2;
    bitmap.pitch = (bitmap.width * bitmap.granularity);
    bitmap.data   = (unsigned char *)VideoBuffer;
    bitmap.viewport.w = 256;
    bitmap.viewport.h = 224;
    bitmap.viewport.x = 0x20;
    bitmap.viewport.y = 0x00;
    bitmap.remap = 1;

    system_init();

    genZ_sndbuffsize=1024;    
    if (genZ_enabledSound&2) 
    {
    	audio_init(genZ_sndfreq,genZ_sndbuffsize);
    	audio_fastmode(genZ_enabledSound&8);
    }
    __asm 
    {
  		ldr  r0, = reg_sys
	  	add	 r0,r0,r10
  		str	 r9,[r0]
	  	str	 r10,[r0,#4]
    } 
    
    
    system_reset();	
    
    genZ_frameTime=(frame_end==0xE0?17:20); //hardcoded to NTSC
            
    genZ_frameBeginTime=genZ_getTickCount();
    

	genZ_pause(0);
    return 0;  
}
void genZ_end()
{
      int val;
	 genZ_pause(1);

	  TwGfxUnlockSurface(genZ_RenderingScreen[0],0);
	  TwGfxUnlockSurface(genZ_RenderingScreen[1],0);

	  if (genZ_RenderingScreen[0]) TwGfxFreeSurface(genZ_RenderingScreen[0]);
	  if (genZ_RenderingScreen[1]) TwGfxFreeSurface(genZ_RenderingScreen[1]);
  
	  TwInputClose(genZ_mInputQueue);  	
	  ljz_free(genZ_mInputFormat);    
/*emu*/

	  
	  if (saveram) 
	  {
	  	save_sram();
	  	ljz_free(saveram);
	  }

 	  system_shutdown();
      error_shutdown();
      
    FtrPtrFree(APPLI_ID, FTR_ROM_ID);                  
      
      

 	if ((genZ_renderingmode)&&(!(genZ_renderingmode&1))) FtrSet(APPLI_ID, FTR_OPT_SMOOTHING_ID,1);
	else FtrSet(APPLI_ID, FTR_OPT_SMOOTHING_ID,0);	

	FtrSet(APPLI_ID, FTR_OPT_SOUND_ID,genZ_enabledSound);
	switch (genZ_sndfreq)
	{
		case 11025:val=2;break;
		case 22050:val=1;break;
		default:val=0;break;
	}
	FtrSet(APPLI_ID, FTR_OPT_SOUNDFREQ_ID,val);
	
	FtrSet(APPLI_ID, FTR_OPT_SCREENMODE_ID,((genZ_renderingmode+1)>>1));
	
	FtrSet(APPLI_ID, FTR_OPT_FSKIP_ID,genZ_frameskip);
	ljz_std_showstat();	
}

void genZ_asmfastrender256(uint8 *screen,uint8 *source,uint32 h,uint32 ofs)
{
	__asm{
		stmfd r13!,{r4-r11}
				
		add	  r1,r1,r3	//ofs
	 loopY:	
		ldmia r1!,{r3-r12}
		stmia r0!,{r3-r12}		
		ldmia r1!,{r3-r12}
		stmia r0!,{r3-r12}		
		ldmia r1!,{r3-r12}
		stmia r0!,{r3-r12}		
		ldmia r1!,{r3-r12}
		stmia r0!,{r3-r12}	//4x 20pix (40bytes/16bpp) : 80pix
		
		ldmia r1!,{r3-r12}
		stmia r0!,{r3-r12}		
		ldmia r1!,{r3-r12}
		stmia r0!,{r3-r12}		
		ldmia r1!,{r3-r12}
		stmia r0!,{r3-r12}		
		ldmia r1!,{r3-r12}
		stmia r0!,{r3-r12}	//4x 20pix (40bytes/16bpp) : 80pix
		//160pix
		ldmia r1!,{r3-r12}
		stmia r0!,{r3-r12}		
		ldmia r1!,{r3-r12}
		stmia r0!,{r3-r12}		
		ldmia r1!,{r3-r12}
		stmia r0!,{r3-r12}		
		ldmia r1!,{r3-r12}
		stmia r0!,{r3-r12}	//4x 20pix (40bytes/16bpp) : 80pix
		//240pix
		
		ldmia r1!,{r3-r10}
		stmia r0!,{r3-r10}	
		//256pix
								
		add	  r1,r1,#512 //(512-256) * 2
		add	  r0,r0,#128 //(320-256) * 2
		
		subs  r2,r2,#1
		bne   loopY
				
		ldmfd r13!,{r4-r11}	
		}		
}

void genZ_asmfastrender320(uint8 *screen,uint8 *source,uint32 h,uint32 ofs)
{
	__asm{
		stmfd r13!,{r4-r11}
				
		add	  r1,r1,r3	//ofs
	 loopY:	
		ldmia r1!,{r3-r12}
		stmia r0!,{r3-r12}		
		ldmia r1!,{r3-r12}
		stmia r0!,{r3-r12}		
		ldmia r1!,{r3-r12}
		stmia r0!,{r3-r12}		
		ldmia r1!,{r3-r12}
		stmia r0!,{r3-r12}	//4x 20pix (40bytes/16bpp) : 80pix
		
		ldmia r1!,{r3-r12}
		stmia r0!,{r3-r12}		
		ldmia r1!,{r3-r12}
		stmia r0!,{r3-r12}		
		ldmia r1!,{r3-r12}
		stmia r0!,{r3-r12}		
		ldmia r1!,{r3-r12}
		stmia r0!,{r3-r12}	//4x 20pix (40bytes/16bpp) : 80pix
		//160pix
		ldmia r1!,{r3-r12}
		stmia r0!,{r3-r12}		
		ldmia r1!,{r3-r12}
		stmia r0!,{r3-r12}		
		ldmia r1!,{r3-r12}
		stmia r0!,{r3-r12}		
		ldmia r1!,{r3-r12}
		stmia r0!,{r3-r12}	//4x 20pix (40bytes/16bpp) : 80pix
		//240pix
		
		ldmia r1!,{r3-r12}
		stmia r0!,{r3-r12}		
		ldmia r1!,{r3-r12}
		stmia r0!,{r3-r12}		
		ldmia r1!,{r3-r12}
		stmia r0!,{r3-r12}		
		ldmia r1!,{r3-r12}
		stmia r0!,{r3-r12}	//4x 20pix (40bytes/16bpp) : 80pix
		//320pix
		
								
		add	  r1,r1,#384 //(512-320) * 2
		
		subs  r2,r2,#1
		bne   loopY
				
		ldmfd r13!,{r4-r11}	
		}		
}


void genZ_updateScreen(void)
{	 
	if (genZ_drawframe)
	{
		TwGfxPointType destLocation = {0, 0};
	    TwGfxSurfaceType* displaySurface;
		TwGfxRectType mDisp;
		int width = (reg[12] & 1) ? 320 : 256;
	    int height = (reg[1] & 8) ? 240 : 224;
	    
		mDisp.x=0;
		mDisp.y=0;
		mDisp.w=width;
		mDisp.h=height;
    	// Get palmos display surface:
	    xassert(TwGfxGetPalmDisplaySurface(genZ_mGfxLib, &displaySurface));	
    
    	if (genZ_specialfskip)
    	{
	    	if (genZ_framecount&1)
        	{
				if (width==256) genZ_asmfastrender256((uint8*)(genZ_curRenderingScreenPtr)+320*2*(height>>1),(uint8*)(VideoBuffer)+512*2*(height>>1),height>>1,bitmap.viewport.x<<1);
	    		if (width==320)	genZ_asmfastrender320((uint8*)(genZ_curRenderingScreenPtr)+320*2*(height>>1),(uint8*)(VideoBuffer)+512*2*(height>>1),height>>1,bitmap.viewport.x<<1);
        	}
        	else
        	{
	        	if (width==256) genZ_asmfastrender256((uint8*)genZ_curRenderingScreenPtr,(uint8*)VideoBuffer,(height>>1),bitmap.viewport.x<<1);
    			if (width==320)	genZ_asmfastrender320((uint8*)genZ_curRenderingScreenPtr,(uint8*)VideoBuffer,(height>>1),bitmap.viewport.x<<1);
	        }
    	}
    	else
    	{
	    	if (width==256) genZ_asmfastrender256((uint8*)genZ_curRenderingScreenPtr,(uint8*)VideoBuffer,height,bitmap.viewport.x<<1);
    		if (width==320)	genZ_asmfastrender320((uint8*)genZ_curRenderingScreenPtr,(uint8*)VideoBuffer,height,bitmap.viewport.x<<1);
    	}
		switch (genZ_renderingmode)
		{    
		case 0:
		{						
			destLocation.x=(480-mDisp.w)>>1;
			destLocation.y=(320-mDisp.h)>>1;
			if (!TwGfxIsSurfaceReady(genZ_curRenderingScreen))
				xassert(TwGfxBitBlt(displaySurface, &destLocation, genZ_curRenderingScreen, &mDisp));
			break;
			}
		case 1:
		{
				TwGfxRectType mDispDest;
				mDispDest.w=width*320/height;
				mDispDest.h=320;				
				mDispDest.x=(480-mDispDest.w)>>1;
				mDispDest.y=0;				
				if (!TwGfxIsSurfaceReady(genZ_curRenderingScreen))
					xassert(TwGfxStretchBlt2(displaySurface, &mDispDest, genZ_curRenderingScreen, &mDisp,twGfxStretchFast));
				break;
			}
		case 2:
		{
				TwGfxRectType mDispDest;
				mDispDest.w=width*320/height;
				mDispDest.h=320;				
				mDispDest.x=(480-mDispDest.w)>>1;
				mDispDest.y=0;	
				if (!TwGfxIsSurfaceReady(genZ_curRenderingScreen))
					xassert(TwGfxStretchBlt2(displaySurface, &mDispDest, genZ_curRenderingScreen, &mDisp,twGfxStretchFast|twGfxStretchSmooth));
				break;
			}
			
	
		case 3:
			{
				TwGfxRectType mDispDest;
				mDispDest.x=0;
				mDispDest.y=0;
				mDispDest.w=480;
				mDispDest.h=320;
				if (!TwGfxIsSurfaceReady(genZ_curRenderingScreen))
					xassert(TwGfxStretchBlt2(displaySurface, &mDispDest, genZ_curRenderingScreen, &mDisp,twGfxStretchFast));
				break;
			}							
		case 4:
			{
				TwGfxRectType mDispDest;
				mDispDest.x=0;
				mDispDest.y=0;
				mDispDest.w=480;
				mDispDest.h=320;
				if (!TwGfxIsSurfaceReady(genZ_curRenderingScreen))
					xassert(TwGfxStretchBlt2(displaySurface, &mDispDest, genZ_curRenderingScreen, &mDisp,twGfxStretchFast|twGfxStretchSmooth));
				break;
			}							
			
		case 5:
			{
				TwGfxRectType mDispDest;
				mDisp.x=(width-240)>>1;
				mDisp.y=(height-160)>>1;
				mDisp.w=240;
				mDisp.h=160;
				mDispDest.x=0;
				mDispDest.y=0;
				mDispDest.w=480;
				mDispDest.h=320;
				if (!TwGfxIsSurfaceReady(genZ_curRenderingScreen))
					xassert(TwGfxStretchBlt2(displaySurface, &mDispDest, genZ_curRenderingScreen, &mDisp,twGfxStretchFast));
				break;
			}	
		case 6:
			{
				TwGfxRectType mDispDest;
				mDisp.x=(width-240)>>1;
				mDisp.y=(height-160)>>1;
				mDisp.w=240;
				mDisp.h=160;
				mDispDest.x=0;
				mDispDest.y=0;
				mDispDest.w=480;
				mDispDest.h=320;
				if (!TwGfxIsSurfaceReady(genZ_curRenderingScreen))
					xassert(TwGfxStretchBlt2(displaySurface, &mDispDest, genZ_curRenderingScreen, &mDisp,twGfxStretchFast|twGfxStretchSmooth));
				break;
			}		
		}
		
//		genZ_curRenderingScreenIndex^=1;
//		genZ_curRenderingScreenPtr=genZ_RenderingScreenPtr[genZ_curRenderingScreenIndex];
//		bitmap.data   = (unsigned char *)genZ_curRenderingScreenPtr;

	}
	
//	frames_displayed++;	
//	genZ_frameBeginTime+=genZ_frameTime*100/genZ_speedpct;	//60Hz

}

void genZ_MainLoop(void)
{
    update_input();
    
    genZ_frameBeginTime+=genZ_frameTime*100/genZ_speedpct;
    
    if ((genZ_frameskip==AUTO_FSKIP)&&(!genZ_fastforward))
	{
		int i,haswaited=0;
		
		//do avoid issue when Tick is looping
		if ((int32)(genZ_frameBeginTime-genZ_getTickCount())>=(int32)(20*genZ_frameTime))
			genZ_frameBeginTime=genZ_getTickCount()+genZ_frameTime*100/genZ_speedpct;					
			
		//while (genZ_getTickCount()<=genZ_frameBeginTime) haswaited=1;
		i=genZ_frameBeginTime-genZ_getTickCount()+1;
        if (i>0) {haswaited=1; SysTaskDelay(i);}
	
		if (haswaited||(genZ_frameskipped>=MAX_FRAME_SKIPPED)) 
		{
			genZ_drawframe=1;
			genZ_frameskipped=0;
		}
		else 
		{
			genZ_drawframe=0;
			genZ_frameskipped++;
		}
	}
	else
	{		
		if (genZ_frameskip==SKEEZIX_FSKIP) genZ_frameskipped=SKEEZIX_FSKIP;
		genZ_frameskipped++;
		
		if (genZ_fastforward)
		{
			if (genZ_frameskipped>GENZ_TURBOFSKIP)
			{
				genZ_frameskipped=0;
				genZ_drawframe=1;
			}
			else genZ_drawframe=0;
		}
		else
		{
			if (genZ_frameskipped>genZ_frameskip)
			{
				genZ_frameskipped=0;
				genZ_drawframe=1;				
			
			}
			else genZ_drawframe=0;
		}
		
		if ((!genZ_fastforward)&&(genZ_drawframe||(genZ_enabledSound&2)))
	    {	
			//do avoid issue when Tick is looping
			int i;
			if ((int32)(genZ_frameBeginTime-genZ_getTickCount())>=(int32)(20*genZ_frameTime))
				genZ_frameBeginTime=genZ_getTickCount()+genZ_frameTime*100/genZ_speedpct;					
				
			//while (genZ_getTickCount()<=genZ_frameBeginTime) ;		
			i=genZ_frameBeginTime-genZ_getTickCount()+1;
	        if (i>0) {SysTaskDelay(i);}
		}
	}


    if(!system_frame(!genZ_drawframe)) 
    {
    	menu_message("lock");
    	menu_waitPen(NULL,NULL,1);
    	system_reset();
    }
	genZ_updateScreen();
	
	
	genZ_framecount++;
}

void genZ_pause(int pause)
{	
   if (pause)
	{
		if (genZ_enabledSound&2) 
		{
			SndStreamStop(snd_stream);
			SndStreamDelete(snd_stream);
		}
	}
	else
	{
		genZ_frameBeginTime=genZ_getTickCount()+genZ_frameTime*100/genZ_speedpct;
		if (genZ_enabledSound&2) 
		{
			SndStreamCreate(&snd_stream,sndOutput,genZ_sndfreq,sndInt16Little,sndStereo,genZ_sndCallBack,&reg_sys,genZ_sndbuffsize);
			SndStreamStart(snd_stream);		
	}	}
}


void genZ_changefskip(void)
{
	char tmp[64];
	int i;
	
	genZ_frameskip++;
	if (genZ_frameskip>SKEEZIX_FSKIP) genZ_frameskip=0;
	
	genZ_specialfskip=0;
	if (genZ_frameskip==SKEEZIX_FSKIP) genZ_specialfskip=1;
	
	if (genZ_frameskip==AUTO_FSKIP) sprintf(tmp,"Frameskip : AUTO");
	else if (genZ_frameskip==SKEEZIX_FSKIP) sprintf(tmp,"Frameskip : Skeezix");
	else sprintf(tmp,"Frameskip : %d",genZ_frameskip);	
	//DrawMessage(tmp,0);
	menu_message(tmp);
	
	i=genZ_getTickCount();
	while ((genZ_getTickCount()-i)<300)
	{
	}
	genZ_frameBeginTime=genZ_getTickCount()+genZ_frameTime*100/genZ_speedpct;
}
void genZ_changerendering(void)
{	
	genZ_renderingmode++;
	if (genZ_renderingmode==7) genZ_renderingmode=0;

	menu_message("Changing screen mode");
	genZ_frameBeginTime=genZ_getTickCount();
	while ((genZ_getTickCount()-genZ_frameBeginTime)<300)
	{
	}
	genZ_frameBeginTime=genZ_getTickCount()+genZ_frameTime*100/genZ_speedpct;
}


void genZ_updatevpad(unsigned int val)
{
	genZ_padState=val;
}

/****************************************************************************/

int genZ_ShowMenuGFXConfig(uint8 *savescr)
{
  uint8 *scr;
  int x,y;
  int redraw;  
  int result;
  int scrmode,smoothing,fskipval;
  RectangleType *item_rect;
  item_rect=(RectangleType*)ljz_malloc(12*sizeof(RectangleType));
  char tmp[16];
  

	
  if (!genZ_renderingmode) {scrmode=0;smoothing=0;}
  else {scrmode=(genZ_renderingmode+1)>>1;smoothing=!(genZ_renderingmode&1);}
	
  fskipval=genZ_frameskip;
	    
  redraw=1;         
  for (;;)
  {  
  	  if (redraw)
  	  {
  	  	redraw=0;
  	  	scr=WinScreenLock(winLockCopy);  
	    if (scr&&savescr) memcpy(scr,savescr,480*320*2);	    
		//if (scr) WinScreenUnlock();
  	  	RGBColorType fg;
	  	WinPushDrawState ();
  	  	WinSetDrawMode(winOverlay);
    	fg.r=50;fg.g=200;fg.b=50;
	    WinSetTextColorRGB(&fg,NULL);
	    WinSetForeColorRGB(&fg,NULL);
	    FntSetFont(boldFont);
		WinPaintChars(LJZ_OSD_TITLE,strlen(LJZ_OSD_TITLE),10,10);	
		WinPopDrawState();
		
		menu_button(310,20,(char*)"BACK",6,&item_rect[0],2);
	  
	  	menu_button(10,80,(char*)"FRAMESKIP",0,&item_rect[1],2);	  
	  	menu_bar(200,80,fskipval,11,&item_rect[2]);
	  	if (fskipval<AUTO_FSKIP) sprintf(tmp,"%d",fskipval);
	  	else if (fskipval==AUTO_FSKIP) sprintf(tmp,"AUTO");
	  	else sprintf(tmp,"Skeezix");
	  	
	  	menu_messageRaw(320,80,tmp);
	  
	  	menu_button(10,130,(char*)"SCREENMODE",0,&item_rect[3],2);
	  	menu_button(200,130,(char*)"1:1",(scrmode==0?2:1),&item_rect[4],0);
	  	menu_button(250,130,(char*)"4:3",(scrmode==1?2:1),&item_rect[5],0);
	  	menu_button(300,130,(char*)"FULL",(scrmode==2?2:1),&item_rect[6],0);
	  	menu_button(350,130,(char*)"ZOOM",(scrmode==3?2:1),&item_rect[7],1);
	  
	  	menu_button(10,180,(char*)"SMOOTHING",0,&item_rect[8],2);
	  	menu_button(200,180,(char*)"OFF",(smoothing?1:2),&item_rect[9],0);
	  	menu_button(250,180,(char*)"ON",(smoothing?2:1),&item_rect[10],0);	  	  	  
	  	
	  	if (scr) WinScreenUnlock();
	
  	  }
	  menu_waitPen(&x,&y,0);
	  
	  
	  //BACK	  
	  if (TEST_PEN_MENU(x,y,item_rect[0])) {result=0;break;}
	  //FSKIP
	  if (TEST_PEN_MENU(x,y,item_rect[2])) 
	  {
	  	fskipval=(x-10-item_rect[2].topLeft.x)*11/100;
	  	if (fskipval<0) fskipval=0;
	  	if (fskipval>11) fskipval=11;
	  	redraw=1;
	  }
	  //SCREEN MODES
	  if (TEST_PEN_MENU(x,y,item_rect[4])) 
	  {
	  	  scrmode=0;
	  	  redraw=1;
	  }
	  if (TEST_PEN_MENU(x,y,item_rect[5])) 
	  {
	  	  scrmode=1;
	  	  redraw=1;
	  }
	  if (TEST_PEN_MENU(x,y,item_rect[6])) 
	  {
	  	  scrmode=2;
	  	  redraw=1;
	  }
	  if (TEST_PEN_MENU(x,y,item_rect[7])) 
	  {
	  	  scrmode=3;
	  	  redraw=1;
	  }
	  //SMOOTHING
	  if (TEST_PEN_MENU(x,y,item_rect[9])&&scrmode) 
	  {
	  	  smoothing=0;
	  	  redraw=1;
	  }
	  if (TEST_PEN_MENU(x,y,item_rect[10])&&scrmode) 
	  {
	  	  smoothing=1;
	  	  redraw=1;
	  }
  }  
  
  genZ_specialfskip=0;
  genZ_frameskip=fskipval;
  if (genZ_frameskip==SKEEZIX_FSKIP) genZ_specialfskip=1;  

  if (!scrmode) genZ_renderingmode=0;
	else 
	if (scrmode==1) genZ_renderingmode=(smoothing?2:1);
	else
	if (scrmode==2) genZ_renderingmode=(smoothing?4:3);
	else
	genZ_renderingmode=(smoothing?6:5);
	
  menu_waitPen(NULL,NULL,2);
  
  ljz_free(item_rect);
  return result;
}



int genZ_ShowMenuSNDConfig(uint8 *savescr)
{
  uint8 *scr;
  int x,y;
  int redraw;  
  int result;
  int vol;
  int bassboost;
  RectangleType *item_rect;
  char tmp[16];
  item_rect=(RectangleType*)ljz_malloc(15*sizeof(RectangleType));
  
  vol=TwSndGetVolume();
  bassboost=TwSndGetBassBoost();
	    
  redraw=1;         
  for (;;)
  {  
  	  if (redraw)
  	  {
  	  	redraw=0;
  	  	scr=WinScreenLock(winLockCopy);  
	    if (scr&&savescr) memcpy(scr,savescr,480*320*2);	    
//		if (scr) WinScreenUnlock();
  	  	RGBColorType fg;
	  	WinPushDrawState ();
  	  	WinSetDrawMode(winOverlay);
    	fg.r=50;fg.g=200;fg.b=50;
	    WinSetTextColorRGB(&fg,NULL);
	    WinSetForeColorRGB(&fg,NULL);
	    FntSetFont(boldFont);
		WinPaintChars(LJZ_OSD_TITLE,strlen(LJZ_OSD_TITLE),10,10);	
		WinPopDrawState();
  	
  		
  	  	  	  	  	  	
	  menu_button(310,20,(char*)"BACK",6,&item_rect[0],2);	  
	  menu_button(10,80,(char*)"SOUND",0,&item_rect[1],2);
	  menu_button(200,80,(char*)"OFF",(genZ_enabledSound==0?2:1),&item_rect[2],0);
	  menu_button(250,80,(char*)"SIM",(genZ_enabledSound&1?2:1),&item_rect[3],0);	  
	  menu_button(300,80,(char*)"ON",(genZ_enabledSound&2?2:1),&item_rect[4],0);	 
	  menu_button(350,80,(char*)"Z80",(genZ_enabledSound&4?2:1),&item_rect[5],0);	  
	  menu_button(400,80,(char*)"FST",(genZ_enabledSound&8?2:1),&item_rect[6],0);	 
	  
/*	  menu_button(10,120,(char*)"VOLUME",0,&item_rect[7],2);	  
	  menu_bar(200,120,vol,64,&item_rect[8]);
	  sprintf(tmp,"%d%",vol*100/64);	  	
	  	menu_messageRaw(320,120,tmp);*/
	  
	  menu_button(10,120,(char*)"BASS BOOST",0,&item_rect[9],2);
	  menu_button(200,120,(char*)"OFF",(bassboost==0?2:1),&item_rect[10],0);
	  menu_button(300,120,(char*)"ON",(bassboost?2:1),&item_rect[11],0);	 
		if (scr) WinScreenUnlock();	  
  	  }
	  menu_waitPen(&x,&y,0);
	  
	  
	  //BACK	  
	  if (TEST_PEN_MENU(x,y,item_rect[0])) {result=0;break;}	  
	  //SOUND OFF
	  if (TEST_PEN_MENU(x,y,item_rect[2])) 
	  {
	  	  if (genZ_enabledSound)
	  	  {
	  	  	if (snd.enabled) audio_shutdown();
		    snd.enabled=0;
			emul_z80=0;
	  	  }
		  genZ_enabledSound=0;		  
		  redraw=1;
	  }
	  //SOUND SIM
	  if (TEST_PEN_MENU(x,y,item_rect[3])) 
	  {
	  	  if (!(genZ_enabledSound&4)) 
	  	  {	
		  	 if (menu_confirm("RESET needed. Continue ?"))
	  	  	 {
	  	  	 	if (snd.enabled) audio_shutdown();
			    snd.enabled=0;
				emul_z80=1;
	  	  	 	genZ_enabledSound&=~3;
  		  	  	genZ_enabledSound|=1|4;
  		  	  	system_reset();
	  	  	 }
	  	  	
	  	  }
	  	  else
	  	  {
		  	  genZ_enabledSound&=~3;
  		  	  genZ_enabledSound|=1|4;
  		  }
	  	  redraw=1;
	  }
	  //SOUND ON
	  if (TEST_PEN_MENU(x,y,item_rect[4])) 
	  {
	  	  if (!(genZ_enabledSound&2))
	  	  {
	  	  	audio_init(genZ_sndfreq,genZ_sndbuffsize);
	  	  }	  	  
	  	  genZ_enabledSound&=~3;
		  genZ_enabledSound|=2;
		  redraw=1;
	  }
	  //ZX80 on/off
	  if (TEST_PEN_MENU(x,y,item_rect[5])) 
	  {
	  	if (!(genZ_enabledSound&4))
	  	{
	  		if (menu_confirm("RESET needed. Continue ?"))
	  	  	{	  	  	 	
				emul_z80=1;
  		  	  	genZ_enabledSound|=4;
  		  	  	system_reset();
	  	  	}	  	  	
	  	}
	  	else
	  	{
	  	 	emul_z80=0;
  		  	genZ_enabledSound&=~4;
	  	}
		redraw=1;
	  }
	  
	  //fast on/off
	  if (TEST_PEN_MENU(x,y,item_rect[6])) 
	  {
	    if (!(genZ_enabledSound&8))
	    {
		  	if (!(genZ_enabledSound&2))
		  	{
	  			audio_init(genZ_sndfreq,genZ_sndbuffsize);  		  		
	  			genZ_enabledSound&=~3;
	  			genZ_enabledSound|=2;
			}
		  	genZ_enabledSound|=8;		  	
		}
		else
		{
			genZ_enabledSound&=~8;
		}
		audio_fastmode(genZ_enabledSound&8);
		redraw=1;
	  }
	  
	  //VOLUME
/*	  if (TEST_PEN_MENU(x,y,item_rect[8])) 
	  {
	  	vol=(x-item_rect[8].topLeft.x)*64/100;
	  	if (vol>64) vol=64;
	  	redraw=1;
	  }*/
	  //BASSBOOST OFF
	  if (TEST_PEN_MENU(x,y,item_rect[9])) 
	  {
	  	bassboost=0;
	  	redraw=1;
	  }
	  //BASSBOOST ON
	  if (TEST_PEN_MENU(x,y,item_rect[10])) 
	  {
	  	bassboost=1;
	  	redraw=1;
	  }
  }  
  
  TwSndSetVolume(vol);
  TwSndSetBassBoost(bassboost);
  
  menu_waitPen(NULL,NULL,2);
  
  ljz_free(item_rect);
  return result;
}


int genZ_ShowMenuOTHERConfig(uint8 *savescr)
{
  uint8 *scr;
  int x,y;
  int redraw;  
  int result;
  RectangleType *item_rect;
  char tmp[16];
  item_rect=(RectangleType*)ljz_malloc(21*sizeof(RectangleType));
  
  redraw=1;         
  for (;;)
  {  
  	  if (redraw)
  	  {
  	  	redraw=0;
  	  	scr=WinScreenLock(winLockCopy);  
	    if (scr&&savescr) memcpy(scr,savescr,480*320*2);	    
  	  	RGBColorType fg;
	  	WinPushDrawState ();
  	  	WinSetDrawMode(winOverlay);
    	fg.r=50;fg.g=200;fg.b=50;
	    WinSetTextColorRGB(&fg,NULL);
	    WinSetForeColorRGB(&fg,NULL);
	    FntSetFont(boldFont);
		WinPaintChars(LJZ_OSD_TITLE,strlen(LJZ_OSD_TITLE),10,10);	
		WinPopDrawState();
  	
  		
  	  	  	  	  	  	
	  menu_button(310,20,(char*)"BACK",6,&item_rect[0],2);	  
	  menu_button(10,80,(char*)"AUTOFIRE A",0,&item_rect[1],2);
	  menu_button(200,80,(char*)"OFF",(genZ_autofireA==0?2:1),&item_rect[2],0);
	  menu_button(250,80,(char*)"SLW",(genZ_autofireA==1?2:1),&item_rect[3],0);
	  menu_button(300,80,(char*)"MED",(genZ_autofireA==2?2:1),&item_rect[4],0);
	  menu_button(350,80,(char*)"FST",(genZ_autofireA==3?2:1),&item_rect[5],0);
	  
	  menu_button(10,120,(char*)"AUTOFIRE B",0,&item_rect[6],2);
	  menu_button(200,120,(char*)"OFF",(genZ_autofireB==0?2:1),&item_rect[7],0);
	  menu_button(250,120,(char*)"SLW",(genZ_autofireB==1?2:1),&item_rect[8],0);
	  menu_button(300,120,(char*)"MED",(genZ_autofireB==2?2:1),&item_rect[9],0);
	  menu_button(350,120,(char*)"FST",(genZ_autofireB==3?2:1),&item_rect[10],0);
	  
	  menu_button(10,160,(char*)"AUTOFIRE C",0,&item_rect[11],2);
	  menu_button(200,160,(char*)"OFF",(genZ_autofireC==0?2:1),&item_rect[12],0);
	  menu_button(250,160,(char*)"SLW",(genZ_autofireC==1?2:1),&item_rect[13],0);
	  menu_button(300,162,(char*)"MED",(genZ_autofireC==2?2:1),&item_rect[14],0);
	  menu_button(350,160,(char*)"FST",(genZ_autofireC==3?2:1),&item_rect[15],0);
	  
	  menu_button(10,200,(char*)"EMU SPEED",0,&item_rect[16],2);	  
	  menu_bar(200,200,genZ_speedpct,100,&item_rect[17]);
	  sprintf(tmp,"%d%",genZ_speedpct);
    	menu_messageRaw(320,200,tmp);
    
      menu_button(10,240,(char*)"CPU CORE",0,&item_rect[18],2);
 	  menu_button(200,240,(char*)"C",(genZ_cpu_cyclone==0?2:1),&item_rect[19],0);
	  menu_button(250,240,(char*)"ASM",(genZ_cpu_cyclone==1?2:1),&item_rect[20],0);
	  
		if (scr) WinScreenUnlock();	  
  	  }
	  menu_waitPen(&x,&y,0);
	  
	  
	  //BACK	  
	  if (TEST_PEN_MENU(x,y,item_rect[0])) {result=0;break;}	  
	  //AUTOFIRE A
	  if (TEST_PEN_MENU(x,y,item_rect[2])) 
	  {
		  genZ_autofireA=0;		  
		  redraw=1;
	  }
	  if (TEST_PEN_MENU(x,y,item_rect[3])) 
	  {
		  genZ_autofireA=1;
		  redraw=1;
	  }
	  if (TEST_PEN_MENU(x,y,item_rect[4])) 
	  {
		  genZ_autofireA=2;
		  redraw=1;
	  }
	  if (TEST_PEN_MENU(x,y,item_rect[5])) 
	  {
		  genZ_autofireA=3;
		  redraw=1;
	  }
	  //AUTOFIRE B
	  if (TEST_PEN_MENU(x,y,item_rect[7])) 
	  {
		  genZ_autofireB=0;		  
		  redraw=1;
	  }
	  if (TEST_PEN_MENU(x,y,item_rect[8])) 
	  {
		  genZ_autofireB=1;
		  redraw=1;
	  }
	  if (TEST_PEN_MENU(x,y,item_rect[9])) 
	  {
		  genZ_autofireB=2;
		  redraw=1;
	  }
	  if (TEST_PEN_MENU(x,y,item_rect[10])) 
	  {
		  genZ_autofireB=3;
		  redraw=1;
	  }
	  //AUTOFIRE C
	  if (TEST_PEN_MENU(x,y,item_rect[12])) 
	  {
		  genZ_autofireC=0;		  
		  redraw=1;
	  }
	  if (TEST_PEN_MENU(x,y,item_rect[13])) 
	  {
		  genZ_autofireC=1;
		  redraw=1;
	  }
	  if (TEST_PEN_MENU(x,y,item_rect[14])) 
	  {
		  genZ_autofireC=2;
		  redraw=1;
	  }
	  if (TEST_PEN_MENU(x,y,item_rect[15])) 
	  {
		  genZ_autofireC=3;
		  redraw=1;
	  }
	  //SPEED
	  if (TEST_PEN_MENU(x,y,item_rect[17])) 
	  {
	  	genZ_speedpct=(x-10-item_rect[17].topLeft.x);
	  	if (genZ_speedpct>100) genZ_speedpct=100;
	  	if (genZ_speedpct<10) genZ_speedpct=10;
	  	redraw=1;
	  }
	  //CPU CORE
	  if (TEST_PEN_MENU(x,y,item_rect[19])) 
	  {
		  if (menu_confirm("RESET needed. Continue ?")) 
		  {
			genZ_cpu_cyclone=0;		  
		  	system_reset();
		  }
		  
		  redraw=1;
	  }
	  if (TEST_PEN_MENU(x,y,item_rect[20])) 
	  {
		  if (menu_confirm("RESET needed. Continue ?")) 
		  {
			genZ_cpu_cyclone=1;
		  	system_reset();
		  }
		  redraw=1;
	  }
  }  
  
  menu_waitPen(NULL,NULL,2);
  
  ljz_free(item_rect);
  return result;
}


int genZ_ShowMenuConfig(uint8 *savescr)
{
  uint8 *scr;
  int x,y;
  int redraw;  
  int result;
  RectangleType *item_rect;
  item_rect=(RectangleType*)ljz_malloc(15*sizeof(RectangleType));
	    
  redraw=1;         
  for (;;)
  {  
  	  if (redraw)
  	  {
  	  	redraw=0;
  	  	scr=WinScreenLock(winLockCopy);  
	    if (scr&&savescr) memcpy(scr,savescr,480*320*2);	    
		//if (scr) WinScreenUnlock();
  	  	RGBColorType fg;
	  	WinPushDrawState ();
  	  	WinSetDrawMode(winOverlay);
    	fg.r=50;fg.g=200;fg.b=50;
	    WinSetTextColorRGB(&fg,NULL);
	    WinSetForeColorRGB(&fg,NULL);
	    FntSetFont(boldFont);
		WinPaintChars(LJZ_OSD_TITLE,strlen(LJZ_OSD_TITLE),10,10);	
		WinPopDrawState();
  	
  		
  	  	  	  	  	  	
	  menu_button(310,20,(char*)"BACK",6,&item_rect[0],2);	  
	  menu_button(172,100,(char*)"GFX",0,&item_rect[1],2);
	  menu_button(172,150,(char*)"SOUND",0,&item_rect[2],2);
	  menu_button(172,200,(char*)"OTHER",0,&item_rect[3],2);
	  menu_button(172,250,(char*)"SND & BRT",9,&item_rect[4],2);
	  if (scr) WinScreenUnlock();
  	  }
	  menu_waitPen(&x,&y,1);
	  
	  
	  //BACK	  
	  if (TEST_PEN_MENU(x,y,item_rect[0])) {result=0;break;}	  
	  //GFX
	  if (TEST_PEN_MENU(x,y,item_rect[1])) 
	  {
	  	genZ_ShowMenuGFXConfig(savescr);
		redraw=1;
	  }
	  //SOUND
	  if (TEST_PEN_MENU(x,y,item_rect[2])) 
	  {
     	genZ_ShowMenuSNDConfig(savescr);
		redraw=1;
	  }
	  //OTHER
	  if (TEST_PEN_MENU(x,y,item_rect[3])) 
	  {
		  genZ_ShowMenuOTHERConfig(savescr);
		  redraw=1;		  
	  }
	  //SND & BRT
	  if (TEST_PEN_MENU(x,y,item_rect[4]))
	  {
	  	  EventType myevent;
	  	  
	  	  for (;;)	  	  
	  	  {
		  	  EvtGetEvent(&myevent,0);
	  	  	  if (myevent.eType==nilEvent) break;			  
		  }
	  	  
	  	  myevent.eType=keyDownEvent;
	  	  myevent.data.keyDown.chr=vchrBrightness;	  	  
	  	  EvtAddEventToQueue(&myevent);
	  	  
	  	  for (;;)	  	  
	  	  {
		  	  EvtGetEvent(&myevent,100);
	  	  	  if (myevent.eType==nilEvent) break;
			  if (myevent.eType==keyDownEvent)
			  {
			  	if (myevent.data.keyDown.chr!=vchrLaunch)
			  	{SysHandleEvent(&myevent);}
			  }			  
		  }
		  redraw=1;
	  }
  }  
  ljz_free(item_rect);
  return result;
}



int genZ_ShowMenuSave(uint8 *savescr)
{
  uint8 *scr;
  int x,y;
  int redraw;  
  int result;
  RectangleType *item_rect;
  item_rect=(RectangleType*)ljz_malloc(15*sizeof(RectangleType));
  memset((char*)item_rect,0,15*sizeof(RectangleType));
  int slot[4];  
  redraw=1;         
  for (;;)
  {  
  	  if (redraw)
  	  {
  	  	redraw=0;
  	  	scr=WinScreenLock(winLockCopy);  
	    if (scr&&savescr) memcpy(scr,savescr,480*320*2);	    
		//if (scr) WinScreenUnlock();
  	  	RGBColorType fg;
	  	WinPushDrawState ();
  	  	WinSetDrawMode(winOverlay);
    	fg.r=50;fg.g=200;fg.b=50;
	    WinSetTextColorRGB(&fg,NULL);
	    WinSetForeColorRGB(&fg,NULL);
	    FntSetFont(boldFont);
		WinPaintChars(LJZ_OSD_TITLE,strlen(LJZ_OSD_TITLE),10,10);	
		WinPopDrawState();
  	
  		
  	  	  	  	  	  	
	  FILE *savefile;
  	 
  	  slot[0]=slot[1]=slot[2]=slot[3]=0;
  	  for (int i=0;i<4;i++) 
  	  if (!state_load(i,1)) 
  	  	{
      		slot[i]=1;
      
	  		struct tm *localTime;
			time_t long_time;
			char str[255];		
			GetStateFileName(i,str);
			long_time=__getdate_file(str);		
			localTime = localtime(&long_time);
			//date is an ofs from 01 Jan 1900 00:00:00			
			sprintf(str,"%02d %s %02d %02d:%02d",
			localTime->tm_mday,str_month[localTime->tm_mon],localTime->tm_year-100,localTime->tm_hour,localTime->tm_min);
			if (i) 
			{
				menu_messageRaw(300,50+i*30,str);
				menu_button(160,50+i*30,"X",5,&item_rect[6+i],3);
			}
			else 
			{
				menu_messageRaw(300,170,str);
				menu_button(160,170,"X",5,&item_rect[6+i],3);
			}
	  	}	  	  	
  	  	  	  	  	  	
	  menu_button(310,20,(char*)"BACK",6,&item_rect[0],2);
	  menu_button(10,80,(char*)"SAVE",0,&item_rect[1],2);
	  menu_button(200,80,(char*)"SLOT 1",(slot[1]?2:1),&item_rect[2],1);
	  menu_button(200,110,(char*)"SLOT 2",(slot[2]?2:1),&item_rect[3],1);
	  menu_button(200,140,(char*)"SLOT 3",(slot[3]?2:1),&item_rect[4],1);
	  menu_button(200,170,(char*)"QUICK",(slot[0]?2:1),&item_rect[5],1);
	  if (scr) WinScreenUnlock();
  	  }
	  menu_waitPen(&x,&y,1);
	  
	  if (TEST_PEN_MENU(x,y,item_rect[6])) 
	  {
	  	if (menu_confirm("Delete QUICKSLOT?"))
	  	{
		  	char str[255];		
			GetStateFileName(0,str);
	  		remove(str);
	  	}
	  	redraw=1;
	  }
	  if (TEST_PEN_MENU(x,y,item_rect[7])) 
	  {
	  	if (menu_confirm("Delete SLOT 1?"))
	  	{
		  	char str[255];		
			GetStateFileName(1,str);
	  		remove(str);
	  	}
	  	redraw=1;
	  }
	  if (TEST_PEN_MENU(x,y,item_rect[8])) 
	  {
	  	if (menu_confirm("Delete SLOT 2?"))
	  	{
		  	char str[255];		
			GetStateFileName(2,str);
	  		remove(str);
	  	}
	  	redraw=1;
	  }
	  if (TEST_PEN_MENU(x,y,item_rect[9])) 
	  {
	  	if (menu_confirm("Delete SLOT 3?"))
	  	{
		  	char str[255];		
			GetStateFileName(3,str);
	  		remove(str);
	  	}
	  	redraw=1;
	  }
	  
	  //BACK	  
	  if (TEST_PEN_MENU(x,y,item_rect[0])) {result=0;break;}
	  //SAVE 0
	  if (TEST_PEN_MENU(x,y,item_rect[2]))
	  {
	  	  if (!slot[1]||(menu_confirm("Overwrite ?")))
		  {state_save(1);result=1;break;}
		  else redraw=1;
		  
	  }
	  //SAVE 1
	  if (TEST_PEN_MENU(x,y,item_rect[3]))
	  {
		  if (!slot[2]||(menu_confirm("Overwrite ?")))
		  {state_save(2);result=1;break;}
		  else redraw=1;
		
	  }
	  //SAVE 2
	  if (TEST_PEN_MENU(x,y,item_rect[4]))
	  {
    	  if (!slot[3]||(menu_confirm("Overwrite ?")))
		  {state_save(3);result=1;break;}
		  else redraw=1;
		
	  }
	  //QUICKSAVE
	  if (TEST_PEN_MENU(x,y,item_rect[5]))
	  {
	  	  if (!slot[0]||(menu_confirm("Overwrite ?")))
		  {state_save(0);result=1;break;}
		  else redraw=1;
		
	  }
  }  
  ljz_free(item_rect);
  return result;
}



int genZ_ShowMenuLoad(uint8 *savescr)
{
  uint8 *scr;
  int x,y;
  int redraw;  
  int result;
   int slot[4];
  RectangleType *item_rect;
  item_rect=(RectangleType*)ljz_malloc(15*sizeof(RectangleType));
  memset((char*)item_rect,0,15*sizeof(RectangleType));
	    
  redraw=1;         
  for (;;)
  {  
  	  if (redraw)
  	  {
  	  	redraw=0;
  	  	scr=WinScreenLock(winLockCopy);  
	    if (scr&&savescr) memcpy(scr,savescr,480*320*2);	    
//		if (scr) WinScreenUnlock();
  	  	RGBColorType fg;
	  	WinPushDrawState ();
  	  	WinSetDrawMode(winOverlay);
    	fg.r=50;fg.g=200;fg.b=50;
	    WinSetTextColorRGB(&fg,NULL);
	    WinSetForeColorRGB(&fg,NULL);
	    FntSetFont(boldFont);
		WinPaintChars(LJZ_OSD_TITLE,strlen(LJZ_OSD_TITLE),10,10);	
		WinPopDrawState();
  	
  	  FILE *savefile;
  	 
  	  slot[0]=slot[1]=slot[2]=slot[3]=0;
      for (int i=0;i<4;i++) 
      	if (!state_load(i,1)) 
      	{
      		slot[i]=1;
      
	  		struct tm *localTime;
			time_t long_time;
			char str[255];		
			GetStateFileName(i,str);
			long_time=__getdate_file(str);		
			localTime = localtime(&long_time);
			//date is an ofs from 01 Jan 1900 00:00:00			
			sprintf(str,"%02d %s %02d %02d:%02d",
			localTime->tm_mday,str_month[localTime->tm_mon],localTime->tm_year-100,localTime->tm_hour,localTime->tm_min);
			if (i) 
			{
				menu_messageRaw(300,50+i*30,str);
				menu_button(160,50+i*30,"X",5,&item_rect[6+i],3);
			}
			else 
			{
				menu_messageRaw(300,170,str);
				menu_button(160,170,"X",5,&item_rect[6+i],3);
			}
	  	}	  	  	
      
	  menu_button(310,20,(char*)"BACK",6,&item_rect[0],2);
	  menu_button(10,80,(char*)"LOAD",0,&item_rect[1],2);
	  menu_button(200,80,(char*)"SLOT 1",(slot[1]?2:3),&item_rect[2],1);
	  menu_button(200,110,(char*)"SLOT 2",(slot[2]?2:3),&item_rect[3],1);
	  menu_button(200,140,(char*)"SLOT 3",(slot[3]?2:3),&item_rect[4],1);
	  menu_button(200,170,(char*)"QUICK",(slot[0]?2:3),&item_rect[5],1);
		if (scr) WinScreenUnlock();	  
  	  }
	  menu_waitPen(&x,&y,1);
	  
	  	  if (TEST_PEN_MENU(x,y,item_rect[6])) 
	  {
	  	if (menu_confirm("Delete QUICKSLOT?"))
	  	{
		  	char str[255];		
			GetStateFileName(0,str);
	  		remove(str);
	  	}
	  	redraw=1;
	  }
	  if (TEST_PEN_MENU(x,y,item_rect[7])) 
	  {
	  	if (menu_confirm("Delete SLOT 1?"))
	  	{
		  	char str[255];		
			GetStateFileName(1,str);
	  		remove(str);
	  	}
	  	redraw=1;
	  }
	  if (TEST_PEN_MENU(x,y,item_rect[8])) 
	  {
	  	if (menu_confirm("Delete SLOT 2?"))
	  	{
		  	char str[255];		
			GetStateFileName(2,str);
	  		remove(str);
	  	}
	  	redraw=1;
	  }
	  if (TEST_PEN_MENU(x,y,item_rect[9])) 
	  {
	  	if (menu_confirm("Delete SLOT 3?"))
	  	{
		  	char str[255];		
			GetStateFileName(3,str);
	  		remove(str);
	  	}
	  	redraw=1;
	  }
	  
	  //BACK	  
	  if (TEST_PEN_MENU(x,y,item_rect[0])) {result=0;break;}
	  //LOAD 0
	  if (TEST_PEN_MENU(x,y,item_rect[2]))
	  {
		  if (slot[1])
		if (!state_load(1,0)) {result=1;break;}
	  }
	  //LOAD 1
	  if (TEST_PEN_MENU(x,y,item_rect[3]))
	  {
		  if (slot[2])
		if (!state_load(2,0)) {result=1;break;}
	  }
	  //LOAD 2
	  if (TEST_PEN_MENU(x,y,item_rect[4]))
	  {
		  if (slot[3])
		if (!state_load(3,0)) {result=1;break;}	  
	  }
	
	  //QUICKLOAD
	  if (TEST_PEN_MENU(x,y,item_rect[5]))
	  {
	  	if (slot[0])
    	if (!state_load(0,0)) {result=1;break;}
	  }
  }  
  ljz_free(item_rect);
  return result;
}


int genZ_ShowSysInfos(uint8 *savescr)
{
  uint8 *scr;
  int x,y;
  int redraw;  
  int result;
  RectangleType *item_rect,my_rect;
  item_rect=(RectangleType*)ljz_malloc(15*sizeof(RectangleType));
	    
  redraw=1;         
  for (;;)
  {  
  	  if (redraw)
  	  {
  	  	redraw=0;
  	  	scr=WinScreenLock(winLockCopy);  
	    if (scr&&savescr) memcpy(scr,savescr,480*320*2);	    
		//if (scr) WinScreenUnlock();
  	  	RGBColorType fg;
	  	WinPushDrawState ();
  	  	WinSetDrawMode(winOverlay);
    	fg.r=50;fg.g=200;fg.b=50;
	    WinSetTextColorRGB(&fg,NULL);
	    WinSetForeColorRGB(&fg,NULL);
	    FntSetFont(boldFont);
		WinPaintChars(LJZ_OSD_TITLE,strlen(LJZ_OSD_TITLE),10,10);	
		WinPopDrawState();
				
		char str[64],strb[32];
		unsigned long maxP,freeP,usedP;		
		VolumeInfoType vinfo;
		VFSVolumeSize(genZ_curVolume,&usedP,&maxP);		
		VFSVolumeGetLabel(genZ_curVolume,strb,16);strb[16]=0;
		menu_button(10,70,"Volume",0,&my_rect,4);		
		sprintf(str,"%s[%d] Used:%dMb/%dMb",strb,genZ_curVolume,usedP>>20,maxP>>20);
		menu_messageRaw(my_rect.topLeft.x+my_rect.extent.x+5,70,str);				
		MemHeapFreeBytes(0,&freeP,&maxP);		
		menu_button(10,100,"RAM",6,&my_rect,4);		
		sprintf(str,"RAM Free:%d Max:%d",freeP,maxP);
		menu_messageRaw(my_rect.topLeft.x+my_rect.extent.x+5,100,str);		
		TwGfxInfoType gfxinfo;
		gfxinfo.size=sizeof(gfxinfo);
		TwGfxGetInfo(genZ_mGfxLib,&gfxinfo);		
		menu_button(10,130,"VRAM",6,&my_rect,4);		
		sprintf(str,"Free:%d Max:%d",gfxinfo.freeAcceleratorMemory,gfxinfo.totalAcceleratorMemory);
		menu_messageRaw(my_rect.topLeft.x+my_rect.extent.x+5,130,str);
						
    		
/*		menu_button(10,160,"ROM",4,&my_rect,4);
		menu_messageRaw(my_rect.topLeft.x+my_rect.extent.x+5,160,rom.name);
		
		sprintf(str,"%dKo RAM-%dKo",mbc.romsize,mbc.ramsize);
		menu_button(10,190,"Size",4,&my_rect,4);
	    menu_messageRaw(my_rect.topLeft.x+my_rect.extent.x+5,190,str);
	    
		menu_button(10,220,"Type",7,&my_rect,4);		
		sprintf(str,"%d Batt:%d Rtc:%d CGB:%d",mbc.type,mbc.batt,rtc.batt,hw.cgb);
		menu_messageRaw(my_rect.topLeft.x+my_rect.extent.x+5,220,str);*/
		
						  	
	  	menu_button(310,20,(char*)"BACK",6,&item_rect[0],2);
		
	  
	  	if (scr) WinScreenUnlock();
  	  }
	  menu_waitPen(&x,&y,1);
	  
	  
	  
	  //BACK	  
	  if (TEST_PEN_MENU(x,y,item_rect[0])) {result=0;break;}
	  
	  
  }  
  ljz_free(item_rect);
  return result;
}



int genZ_ShowMenu()
{
  int x,y;
  int redraw;
  int result;
  int slot0;
  uint8 *scr,*savescr;
  RectangleType *item_rect;
  item_rect=(RectangleType*)ljz_malloc(15*sizeof(RectangleType));  
  redraw=1;
  
  genZ_pause(1);
    
  
   TwGfxSurfaceType* displaySurface;
  // Get palmos display surface:
  xassert(TwGfxGetPalmDisplaySurface(genZ_mGfxLib, &displaySurface));
  while (TwGfxIsSurfaceReady(displaySurface));

  
  savescr=(uint8*)ljz_malloc(320*480*2);  
  scr=WinScreenLock(winLockCopy);  
  if (scr&&savescr)
  {
	  for (x=0;x<320*480;x++)
	  {
  		y=((uint16*)scr)[x];
	  	((uint16*)scr)[x]=((y>>12)<<11)|(((y>>6)&31)<<5)|((y>>1)&15);
	  }
	  memcpy(savescr,scr,480*320*2);
  }
  if (scr) WinScreenUnlock();
            
              
  for (;;)
  {  
  	if (redraw)
  	{  	
	  	slot0=!state_load(0,1);    
  		redraw=0;
		scr=WinScreenLock(winLockCopy);  
	    if (scr&&savescr) memcpy(scr,savescr,480*320*2);	    
//		if (scr) WinScreenUnlock();
  		
  		RGBColorType fg;
	  	WinPushDrawState ();
  	  	WinSetDrawMode(winOverlay);
    	fg.r=50;fg.g=200;fg.b=50;
	    WinSetTextColorRGB(&fg,NULL);
	    WinSetForeColorRGB(&fg,NULL);
	    FntSetFont(boldFont);
		WinPaintChars(LJZ_OSD_TITLE,strlen(LJZ_OSD_TITLE),10,10);	
		WinPopDrawState();
  		
		menu_button(310,20,(char*)"BACK",6,&item_rect[0],2);
		menu_button(60,100,(char*)"SAVE",0,&item_rect[1],2);
		menu_button(280,100,(char*)"QUICK SAVE",(slot0?2:1),&item_rect[2],2);
		menu_button(60,150,(char*)"LOAD",0,&item_rect[3],2);
		menu_button(280,150,(char*)"QUICK LOAD",(slot0?2:3),&item_rect[4],2);
		menu_button(60,200,(char*)"CONFIG",0,&item_rect[5],2);
		menu_button(280,250,(char*)"LAUNCHER",4,&item_rect[6],2);
		
		menu_button(280,200,(char*)"SYS INFOS",7,&item_rect[8],2);
		
		menu_button(60,250,(char*)"RESET GEN.",9,&item_rect[7],2);				
		if (scr) WinScreenUnlock();		
		
  	}
	  menu_waitPen(&x,&y,1);
	  //HOME
	  if (TEST_PEN_MENU(x,y,item_rect[6])) {result=1;break;}
	  //BACK
	  if (TEST_PEN_MENU(x,y,item_rect[0])) {result=0;break;}
	  //CONFIG
	  if (TEST_PEN_MENU(x,y,item_rect[5])) {genZ_ShowMenuConfig(savescr);redraw=1;}
	  //LOAD
	  if (TEST_PEN_MENU(x,y,item_rect[3])) {if (genZ_ShowMenuLoad(savescr)) {result=0;break;} redraw=1;}
	  //SAVE
	  if (TEST_PEN_MENU(x,y,item_rect[1])) {if (genZ_ShowMenuSave(savescr)) {result=0;break;} redraw=1;}
	  //QUICK SAVE
	  if (TEST_PEN_MENU(x,y,item_rect[2]))
	  {
		  state_save(0);
		  result=0;
		  break;
	  }
	  //QUICK LOAD
	  if (TEST_PEN_MENU(x,y,item_rect[4]))
	  {
		  if (slot0)
		    if (!state_load(0,0)) {result=0;break;}
	  }
	  //RESET Genesis
	  if (TEST_PEN_MENU(x,y,item_rect[7])) {if (menu_confirm("RESET GENESIS ?")) {system_reset(); result=0;break;}redraw=1;}
	  
	  //SYS INFOS
	  if (TEST_PEN_MENU(x,y,item_rect[8])) {genZ_ShowSysInfos(savescr); redraw=1;}
	}
  ljz_free(item_rect);
  
  ljz_free(savescr);
            
  genZ_pause(0);
  return result;
}  

