; Convert a VGA video buffer to GP32 format
;  by Jeff F, 2002-Apr-4
;
;  This GCC assembly routine converts an 8-bit VGA style
; video buffer to GP32 8-bit video frame buffer format.
;
;  The default Performance is set to 4. The smallest code
; is obtained by setting this value to 1. The fastest and
; largest code is obtained by setting this value to 80.
; Here are the only valid values for Performance:
;   1,2,4,5,8,10,16,20,40,80
; The percentage gain of speed improvements drops off
; exponentially with each larger Performance setting
; so 4 is probably fine for most cases.
;
; Compile this file by using the following:
;  as -o xlatgp32.o xlatgp32.s
; Then include xlatgp32.o in your link files list.
;
; Reference it from C by using:
;  extern void XlatVGABuffer (unsigned char *src, unsigned char *dest);
;
; Entry:
;  r0 = src, r1 = dest

	    AREA    Xlat, CODE, READONLY

Performance EQU 64
        
		EXPORT XlatNESBufferNTSC
		EXPORT XlatNESBufferPAL
	    EXPORT   XlatNESBufferLineZ	
	    EXPORT   XlatNESBufferLine
	    EXPORT   XlatNESBufferLineZ16
        EXPORT   XlatNESBufferLineZ16B	    
        EXPORT   XlatNESBufferLine16
        
        EXPORT	 XlatNESBufferLineLazy
        EXPORT	 XlatNESBufferLineZLazy
        
        IMPORT	  scalePal2
        IMPORT	  ppuScanline

	
XlatNESBufferNTSC PROC
		stmfd r13!,{r4-r11}
        mov r12,#224
        add r1,r1,#232        
        add r1,r1,#7680
                        
        add r0,r0,#8
        add r0,r0,#2176
XlatLoop1N
        sub r1,r1,#1
        mov r3,r1
        
        

        GBLA	count
count	SETA	0
		WHILE	count < Performance
count	SETA	count+1

        ldr  r4,[r0],#4        
        
        and  r5,r4,#0xff        
         mov  r4,r4,lsr #8
        ldrb r8,[r2,r5]        
         and  r6,r4,#0xff                
          mov  r4,r4,lsr #8        
         ldrb r9,[r2,r6]
          and  r7,r4,#0xff               
           mov  r4,r4,lsr #8                
          ldrb r5,[r2,r7]                      
           ldrb r6,[r2,r4]
                
        strb r8,[r3],#240
         strb r9,[r3],#240
          strb r5,[r3],#240
           strb r6,[r3],#240
        
		WEND
        
        add r0,r0,#16        

        subs r12,r12,#1
        bne XlatLoop1N

        ldmfd r13!,{r4-r11}        
        mov		pc,lr
		ENDP		
        
        
            
XlatNESBufferLine PROC
        stmfd r13!,{r4-r9}
        add   r1,r1,#239
        sub   r3,r1,r3
        add   r3,r3,#7680
        
		GBLA	count
count	SETA	0
		WHILE	count < Performance
count	SETA	count+1
;        ldr  r4,[r0],#4        
        
;        and  r5,r4,#0xff        
;         mov  r4,r4,lsr #8
;        ldrb r8,[r2,r5]        
;         and  r6,r4,#0xff                
;          mov  r4,r4,lsr #8        
;         ldrb r9,[r2,r6]
;          and  r7,r4,#0xff               
;           mov  r4,r4,lsr #8                
;          ldrb r5,[r2,r7]                      
;           ldrb r6,[r2,r4]
                
;        strb r8,[r3],#240
;         strb r9,[r3],#240
;          strb r5,[r3],#240
;           strb r6,[r3],#240
		ldrb  r5,[r0],#1
         ldrb  r6,[r0],#1        
        ldrb r8,[r2,r5]
         ldrb r9,[r2,r6]
         
        ldrb  r7,[r0],#1
         ldrb  r4,[r0],#1         
        ldrb r5,[r2,r7]                      
         ldrb r6,[r2,r4]
                
        strb r8,[r3],#240
         strb r9,[r3],#240
          strb r5,[r3],#240
           strb r6,[r3],#240
        
		WEND
        

        ldmfd r13!,{r4-r9}
        mov		pc,lr
		ENDP


XlatNESBufferLineLazy PROC
        stmfd r13!,{r4-r9}
        
        ldr	  r12, = ppuScanline
        ldr	  r12,[r12]                        
        sub	  r12,r12,r3
        
                
        add   r1,r1,#240
        sub   r3,r1,r3
        add   r1,r3,#7680
        
        b	  XlatLineLazyLoop
        LTORG
XlatLineLazyLoop
        sub   r1,r1,#1
        mov   r3,r1        
        
		GBLA	count
count	SETA	0
		WHILE	count < Performance
count	SETA	count+1
;        ldr  r4,[r0],#4        
        
;        and  r5,r4,#0xff        
;         mov  r4,r4,lsr #8
;        ldrb r8,[r2,r5]        
;         and  r6,r4,#0xff                
;          mov  r4,r4,lsr #8        
;         ldrb r9,[r2,r6]
;          and  r7,r4,#0xff               
;           mov  r4,r4,lsr #8                
;          ldrb r5,[r2,r7]                      
;           ldrb r6,[r2,r4]

        ldrb  r5,[r0],#1
         ldrb  r6,[r0],#1        
        ldrb r8,[r2,r5]
         ldrb r9,[r2,r6]
         
        ldrb  r7,[r0],#1
         ldrb  r4,[r0],#1         
        ldrb r5,[r2,r7]                      
         ldrb r6,[r2,r4]
                
        strb r8,[r3],#240
         strb r9,[r3],#240
          strb r5,[r3],#240
           strb r6,[r3],#240
        
		WEND
        add   r0,r0,#16
        
        subs  r12,r12,#1
        bne	  XlatLineLazyLoop

        ldmfd r13!,{r4-r9}
        mov		pc,lr
		ENDP
		
XlatNESBufferLineZLazy PROC
        stmfd r13!,{r4-r9}
        
        ldr	  r12, = ppuScanline
        ldr	  r12,[r12]                        
        sub	  r12,r12,r3
        
                
        add   r1,r1,#240
        sub   r1,r1,r3
        
        b	  XlatLineLazyZLoop
        LTORG
XlatLineLazyZLoop
        sub   r1,r1,#1
        mov   r3,r1        
        
		GBLA	countZ
countZ	SETA	0
		WHILE	countZ < Performance
countZ	SETA	countZ+1
        ldrb  r5,[r0],#1
         ldrb  r6,[r0],#1        
        ldrb r8,[r2,r5]
         ldrb r9,[r2,r6]
         
        ldrb  r7,[r0],#1
         ldrb  r4,[r0],#1         
        ldrb r5,[r2,r7]                      
         ldrb r6,[r2,r4]
                
        strb r8,[r3],#240
         strb r9,[r3],#240
          strb r5,[r3],#240
           strb r6,[r3],#240
           strb r6,[r3],#240
        
		WEND
        add   r0,r0,#16
        
        subs  r12,r12,#1
        bne	  XlatLineLazyZLoop

        ldmfd r13!,{r4-r9}
        mov		pc,lr
		ENDP
		


        
XlatNESBufferLineZ PROC
        stmfd r13!,{r4-r9}
        
        add   r1,r1,#239
        sub   r3,r1,r3
        
        
        GBLA	count
count	SETA	0
		WHILE	count < Performance
count	SETA	count+1
        ldrb  r5,[r0],#1
         ldrb  r6,[r0],#1        
        ldrb r8,[r2,r5]
         ldrb r9,[r2,r6]
         
        ldrb  r7,[r0],#1
         ldrb  r4,[r0],#1         
        ldrb r5,[r2,r7]                      
         ldrb r6,[r2,r4]
                
        strb r8,[r3],#240
         strb r9,[r3],#240
          strb r5,[r3],#240
           strb r6,[r3],#240
           strb r6,[r3],#240
        
		WEND
        

        ldmfd r13!,{r4-r9}
        mov		pc,lr
		ENDP        		
                
XlatNESBufferLineZ16 PROC
        stmfd r13!,{r4-r12}                
        
        ldr   r11, = 480	
        add   r1,r1,#239
        add   r1,r1,#239
        sub   r3,r1,r3, lsl #1                

        ldr   r10, = scalePal2
        mov   r12,#0
        
        b     ltorg_Z16
        LTORG
ltorg_Z16

        
      
        sub   r0,r0,#1
        ldrb  r4,[r0],#1
        ldrb  r5,[r0]                
        ldrb  r6,[r2,r4]
        ldrb  r7,[r2,r5]
        ldr   r8,[r10,r6,lsl #2]
        ldr   r9,[r10,r7,lsl #2] 
        

        GBLA	count
count	SETA	0
		WHILE	count < Performance
count	SETA	count+1
                         
        GBLA	count2
count2	SETA	0
		WHILE	count2 < 3
count2	SETA	count2+1
        ldrb  r5,[r0],#1
        mov   r12,r8              ;;pix -1
        ldrb  r6,[r2,r5]        
        mov   r8,r9               ;;pix + 0        
        ldr   r9,[r10,r6,lsl #2]  ;;pix + 1                         
        add   r1,r12,r8
        add   r1,r1,r9                        
        add   r5,r8,r1        
        mov   r1,r5,lsr #18  ;;R
        mov   r4,r5,lsr #10
        and   r4,r4,#0x1F    ;;V
        mov   r5,r5,lsr #2
        and   r5,r5,#0x1F    ;;B        
        mov   r6,r1,lsl #11
        orr   r6,r6,r4,lsl #6
        orr   r6,r6,r5,lsl #1                                         
        strh  r6,[r3],r11
		WEND
        mov   r12,r8              ;;pix -1        
        mov   r8,r9               ;;pix + 0        
        ;;pix + 1        is the same                 
        add   r1,r12,r8
        add   r1,r1,r9                        
        add   r5,r8,r1        
        mov   r1,r5,lsr #18  ;;R
        mov   r4,r5,lsr #10
        and   r4,r4,#0x1F    ;;V
        mov   r5,r5,lsr #2
        and   r5,r5,#0x1F    ;;B        
        mov   r6,r1,lsl #11
        orr   r6,r6,r4,lsl #6
        orr   r6,r6,r5,lsl #1                                         
        strh  r6,[r3],r11
        
        ldrb  r5,[r0],#1
        mov   r12,r8              ;;pix -1
        ldrb  r6,[r2,r5]        
        mov   r8,r9               ;;pix + 0        
        ldr   r9,[r10,r6,lsl #2]  ;;pix + 1        
        add   r1,r12,r8
        add   r1,r1,r9                        
        add   r5,r8,r1        
        mov   r1,r5,lsr #18  ;;R
        mov   r4,r5,lsr #10
        and   r4,r4,#0x1F    ;;V
        mov   r5,r5,lsr #2
        and   r5,r5,#0x1F    ;;B        
        mov   r6,r1,lsl #11
        orr   r6,r6,r4,lsl #6
        orr   r6,r6,r5,lsl #1                                         
        strh  r6,[r3],r11                                 
		WEND
        
        ldmfd r13!,{r4-r12}
        mov		pc,lr
end1NLineZ16        
		ENDP
		
        
XlatNESBufferLineZ16B PROC
        stmfd r13!,{r4-r12}                
        
        ldr   r11, = 480	
        add   r1,r1,#239
        add   r1,r1,#239
        sub   r3,r1,r3, lsl #1                

        ldr   r10, = scalePal2
        
        b     ltorg_Z16B
        LTORG
ltorg_Z16B        
                        
        ldrb  r5,[r0]                        
        ldrb  r7,[r2,r5]        
        ldr   r9,[r10,r7,lsl #2] 
        

        GBLA	count
count	SETA	0
		WHILE	count < Performance
count	SETA	count+1
                         
        GBLA	count2
count2	SETA	0
		WHILE	count2 < 3
count2	SETA	count2+1
        ldrb  r5,[r0],#1        
        mov   r8,r9               ;;pix + 0
        ldrb  r6,[r2,r5]                
        ldr   r9,[r10,r6,lsl #2]  ;;pix + 1                                         
        add   r5,r8,r9
        mov   r1,r5,lsr #17  ;;R
        mov   r4,r5,lsr #9
        and   r4,r4,#0x1F    ;;V
        mov   r5,r5,lsr #1
        and   r5,r5,#0x1F    ;;B        
        mov   r6,r1,lsl #11
        orr   r6,r6,r4,lsl #6
        orr   r6,r6,r5,lsl #1                                         
        strh  r6,[r3],r11
		WEND
        mov   r8,r9               ;;pix + 0        
        ;;pix + 1        is the same                         
        add   r5,r8,r9
        mov   r1,r5,lsr #17  ;;R
        mov   r4,r5,lsr #9
        and   r4,r4,#0x1F    ;;V
        mov   r5,r5,lsr #1
        and   r5,r5,#0x1F    ;;B        
        mov   r6,r1,lsl #11
        orr   r6,r6,r4,lsl #6
        orr   r6,r6,r5,lsl #1                                         
        strh  r6,[r3],r11
        
        ldrb  r5,[r0],#1
        mov   r8,r9               ;;pix + 0        
        ldrb  r6,[r2,r5]                
        ldr   r9,[r10,r6,lsl #2]  ;;pix + 1                
        add   r5,r8,r9        
        mov   r1,r5,lsr #17  ;;R
        mov   r4,r5,lsr #9
        and   r4,r4,#0x1F    ;;V
        mov   r5,r5,lsr #1
        and   r5,r5,#0x1F    ;;B        
        mov   r6,r1,lsl #11
        orr   r6,r6,r4,lsl #6
        orr   r6,r6,r5,lsl #1                                         
        strh  r6,[r3],r11                                 
		WEND
        
        ldmfd r13!,{r4-r12}
        mov		pc,lr
end1NLineZ16B
	    ENDP

        
XlatNESBufferLine16 PROC
        stmfd r13!,{r4-r12}                
        
        ldr   r11, = 480	
        add   r1,r1,#239
        add   r1,r1,#239
        ldr   r4, = 15360
        add   r1,r1,r4
        sub   r3,r1,r3, lsl #1                

        ldr   r10, = scalePal2 
        
        b     ltorg_16
        LTORG
ltorg_16
       
                        
        ldrb  r5,[r0]                        
        ldrb  r7,[r2,r5]        
        ldr   r9,[r10,r7,lsl #2] 
        
		GBLA	count
count	SETA	0
		WHILE	count < Performance*4
count	SETA	count+1                         

        ldrb  r5,[r0],#1        
        mov   r8,r9               ;;pix + 0
        ldrb  r6,[r2,r5]                
        ldr   r9,[r10,r6,lsl #2]  ;;pix + 1                                         
        add   r5,r8,r9
        mov   r1,r5,lsr #17  ;;R
        mov   r4,r5,lsr #9
        and   r4,r4,#0x1F    ;;V
        mov   r5,r5,lsr #1
        and   r5,r5,#0x1F    ;;B        
        mov   r6,r1,lsl #11
        orr   r6,r6,r4,lsl #6
        orr   r6,r6,r5,lsl #1                                         
        strh  r6,[r3],r11
		WEND
        
        
        ldmfd r13!,{r4-r12}
        mov		pc,lr
end1NLine16
		ENDP



XlatNESBufferPAL PROC
        stmfd r13!,{r4-r11}
        mov r12,#240
        add r1,r1,#240        
        add r1,r1,#7680                
        
        add r0,r0,#8        
XlatLoop1P
        sub r1,r1,#1
        mov r3,r1
        
        
		GBLA	count
count	SETA	0
		WHILE	count < Performance
count	SETA	count+1                         

        ldr  r4,[r0],#4        
        
        and  r5,r4,#0xff        
         mov  r4,r4,lsr #8
        ldrb r8,[r2,r5]        
         and  r6,r4,#0xff                
          mov  r4,r4,lsr #8        
         ldrb r9,[r2,r6]
          and  r7,r4,#0xff               
           mov  r4,r4,lsr #8                
          ldrb r5,[r2,r7]                      
           ldrb r6,[r2,r4]
                
        strb r8,[r3],#240                
         strb r9,[r3],#240                
          strb r5,[r3],#240        
           strb r6,[r3],#240        
        WEND
        
        add r0,r0,#16        

        subs r12,r12,#1
        bne XlatLoop1P

        ldmfd r13!,{r4-r11}
        mov		pc,lr
		ENDP

		END