
#include "neopop.h"
#include "TLCS900H.h"
#include "mem.h"
#include "gfx.h"
#include "interrupt.h"
#include "Z80_interface.h"
#include "dma.h"


//extern int scanlines_here; //definie dans gp_main.c

//=============================================================================

_u32 timer_hint;
_u32 timer_clock0, timer_clock1, timer_clock2, timer_clock3;
_u8 timer[4];	//Up-counters

BOOL gfx_hack = FALSE;

//=============================================================================

static BOOL h_int = FALSE, timer0, timer2;

//=============================================================================

void interrupt(_u8 index)
{
	push32((&tlcs),get_pc((&tlcs)));
	push16((&tlcs),sr2);

	//Up the IFF
//	if (((sr2 & 0x7000) >> 12) < 7)
// test decal ATTENTION
	if ((sr2 & 0x7000) >> 5)
		setStatusIFF((&tlcs),((sr2 & 0x7000) >> 12) + 1);

	//Access the interrupt vector table to find the jump destination
//	_pc = *(_u32*)(ram + 0x6FB8 + (index * 4));
	set_pc((&tlcs), *(_u32*)(ram + 0x6FB8 + (index << 2)));

}

//=============================================================================

static __inline void gfx_hint(void)
{
	//H_Int / Delayed settings
	if ((ram[0x8009] < SCREEN_HEIGHT-1 || ram[0x8009] == 198))
	{
		gfx_delayed_settings();	//Get delayed settings

		//Allowed?
		if (ram[0x8000] & 0x40)
			h_int = 1;
	}
}

static __inline void gfx_draw(void)
{
	if (frameskip_count == 0)
	{
		//Draw the scanline
		if (ram[0x8009] < SCREEN_HEIGHT)
		{
			if (ram[0x6F95] == 0x10)	gfx_draw_scanline_colour();
			else				gfx_draw_scanline_mono();
		}
	}
}



void updateTimers(_u8 cputicks)
{
	//increment H-INT timer
	timer_hint += cputicks;

	//=======================

	//End of scanline / Start of Next one
	if (timer_hint >= TIMER_HINT_RATE)
	{
		_u8 data;

		// ============= END OF CURRENT SCANLINE =============


		 if (gfx_hack)	{	gfx_draw(); gfx_hint(); }
		 else			{	gfx_hint();	gfx_draw(); }


 
		// ============= START OF NEXT SCANLINE =============
    	ram[0x8009]++;	//Next scanline
  //      if((!scanlines_here) && (ram[0x8009] < SCREEN_HEIGHT)) ram[0x8009]++; // line up

		timer_hint = 0;	//Start of next scanline

		//Comms. Read interrupt
		if ((ram[0xB2] & 1) == 0 && system_comms_poll(&data) &&  (statusIFF((&tlcs)) <= (ram[0x77] & 7)))
		{
		 ram[0x50] = data;
		 interrupt(12); 
		 if (ram[0x007C] == 0x19)		DMA_update(0);
		 else { if (ram[0x007D] == 0x19)	DMA_update(1);
		 else { if (ram[0x007E] == 0x19)	DMA_update(2);
		 else { if (ram[0x007F] == 0x19)	DMA_update(3);	}}}
		}

		//V_Int?
		if (ram[0x8009] == SCREEN_HEIGHT)
		{
         

		 ram[0x8010] = 0x40;	//Character Over / Vblank Status

// ATTENTION ENLVER
// remis la pour des tests
			system_VBL();	//Update the screen

  		 //Vertical Interrupt? (Confirmed IRQ Level)
		 if (statusIFF((&tlcs)) <= 4 && (ram[0x8000] & 0x80))
   	 	 {
          interrupt(5); // VBL
	// 		system_VBL();	//Update the screen
    	  if (ram[0x007C] == 0x0B)		DMA_update(0);
		  else { if (ram[0x007D] == 0x0B)	DMA_update(1);
		  else { if (ram[0x007E] == 0x0B)	DMA_update(2);
		  else { if (ram[0x007F] == 0x0B)	DMA_update(3);	}}}
		 }

		}

		//End of V_Int
		if (ram[0x8009] == 199)	//Last scanline + 1
		{
			ram[0x8009] = 0;
			ram[0x8010] = 0;	//Character Over / Vblank Status
		}
	}

	//=======================

	//Tick the Clock Generator
	timer_clock0 += cputicks;
	timer_clock1 += cputicks;
	
	timer0 = FALSE;	//Clear the timer0 tick, for timer1 chain mode.

	//=======================

	//Run Timer 0 (TRUN)?
	if ((ram[0x20] & 0x01))
	{
		//T01MOD
		switch(ram[0x24] & 0x03)
		{
		case 0:	if (h_int)	//Horizontal interrupt trigger
				{
					timer[0]++;
					timer_clock0 = 0;
					h_int = FALSE;	// Stop h_int remaining active
				}
				break;

		case 1:	if (timer_clock0 >= TIMER_T1_RATE)
				{
					timer[0]++;
					timer_clock0 = 0;
				}
				break;

		case 2:	if (timer_clock0 >= TIMER_T4_RATE)
				{
					timer[0]++;
					timer_clock0 = 0;
				}
				break;

		case 3:	if (timer_clock0 >= TIMER_T16_RATE)
				{
					timer[0]++;
					timer_clock0 = 0;
				}
				break;
		}

		//Threshold check
		if (ram[0x22] && timer[0] >= ram[0x22])
		{
			timer[0] = 0;
			timer0 = TRUE;

			if (statusIFF((&tlcs)) <= (ram[0x73] & 0x7))
				interrupt(7); // Timer 0 Int.

			if (ram[0x007C] == 0x10)		DMA_update(0);
			else { if (ram[0x007D] == 0x10)	DMA_update(1);
			else { if (ram[0x007E] == 0x10)	DMA_update(2);
			else { if (ram[0x007F] == 0x10)	DMA_update(3);	}}}
		}
	}


	//=======================

	//Run Timer 1 (TRUN)?
	if ((ram[0x20] & 0x02))
	{
		//T23MOD
		switch((ram[0x24] & 0x0C) >> 2)
		{
		case 0:	if (timer0)	//Timer 0 chain mode.
				{
					timer[1]++;
					timer_clock1 = 0;
				}
				break;

		case 1:	if (timer_clock1 >= TIMER_T1_RATE)
				{
					timer[1]++;
					timer_clock1 = 0;
				}
				break;

		case 2:	if (timer_clock1 >= TIMER_T16_RATE)
				{
					timer[1]++;
					timer_clock1 = 0;
				}
				break;

		case 3:	if (timer_clock1 >= TIMER_T256_RATE)
				{
					timer[1]++;
					timer_clock1 = 0;
				}
				break;
		}

		//Threshold check
		if (ram[0x23] && timer[1] >= ram[0x23])
		{
			timer[1] = 0;

			if (statusIFF((&tlcs)) <= ((ram[0x73] & 0x70) >> 4))
				interrupt(8); // Timer 1 Int.

			if (ram[0x007C] == 0x11)		DMA_update(0);
			else { if (ram[0x007D] == 0x11)	DMA_update(1);
			else { if (ram[0x007E] == 0x11)	DMA_update(2);
			else { if (ram[0x007F] == 0x11)	DMA_update(3);	}}}
		}
	}

	//=======================

	//Tick the Clock Generator
	timer_clock2 += cputicks;
	timer_clock3 += cputicks;

	timer2 = FALSE;	//Clear the timer2 tick, for timer3 chain mode.

	//=======================
	
	
	//Run Timer 2 (TRUN)?
	if ((ram[0x20] & 0x04))
	{
		//T23MOD
		switch(ram[0x28] & 0x03)
		{
		case 0:	// -
				break;
/////////// ATENTION ATTENTION
//		case 1:	if (timer_clock2 >= TIMER_T1_RATE)	//HACK - Fixes DAC

//		case 1:	if (timer_clock2 >= TIMER_T1_RATE)	//HACK - Fixes DAC
		case 1:	if (timer_clock2 >= 56)	//HACK - Fixes DAC
				{
					timer[2]++;
					timer_clock2 = 0;
				}
				break;

		case 2:	if (timer_clock2 >= TIMER_T4_RATE)
				{
					timer[2]++;
					timer_clock2 = 0;
				}
				break;

		case 3:	if (timer_clock2 >= TIMER_T16_RATE)
				{
					timer[2]++;
					timer_clock2 = 0;
				}
				break;
		}

		//Threshold check
		if (ram[0x26] && timer[2] >= ram[0x26])
		{
			timer[2] = 0;
			timer2 = TRUE;

			if (statusIFF((&tlcs)) <= ((ram[0x74] & 0x07)))
				interrupt(9);	// Timer 2 Int.

			if (ram[0x007C] == 0x12)		DMA_update(0);
			else { if (ram[0x007D] == 0x12)	DMA_update(1);
			else { if (ram[0x007E] == 0x12)	DMA_update(2);
			else { if (ram[0x007F] == 0x12)	DMA_update(3);	}}}
		}
	}

	//=======================

	//Run Timer 3 (TRUN)?
	if ((ram[0x20] & 0x08))
	{
		//T23MOD
		switch((ram[0x28] & 0x0C) >> 2)
		{
		case 0:	if (timer2)	//Timer 2 chain mode.
				{
					timer[3]++;
					timer_clock3 = 0;
				}
				break;

		case 1:	if (timer_clock3 >= TIMER_T1_RATE)
				{
					timer[3]++;
					timer_clock3 = 0;
				}
				break;

		case 2:	if (timer_clock3 >= TIMER_T16_RATE)
				{
					timer[3]++;
					timer_clock3 = 0;
				}
				break;

		case 3:	if (timer_clock3 >= TIMER_T256_RATE)
				{
					timer[3]++;
					timer_clock3 = 0;
				}
				break;
		}

		//Threshold check
		if (ram[0x27] && timer[3] >= ram[0x27])
		{
			timer[3] = 0;

//////// ATTENTION Z80 ICI !!!!
			Z80_irq();
//      z80_set_irq_line(0,CLEAR_LINE);


			if (statusIFF((&tlcs)) <= ((ram[0x74] & 0x70) >> 4))
				interrupt(10); // Timer 3 Int.

			if (ram[0x007C] == 0x13)		DMA_update(0);
			else { if (ram[0x007D] == 0x13)	DMA_update(1);
			else { if (ram[0x007E] == 0x13)	DMA_update(2);
			else { if (ram[0x007F] == 0x13)	DMA_update(3);	}}}
		}
	}

	//=======================
}

//=============================================================================

void reset_timers(void)
{
	timer_hint = 0;

	timer[0] = 0;
	timer[1] = 0;
	timer[2] = 0;
	timer[3] = 0;
	
	timer_clock0 = 0;
	timer_clock1 = 0;
	timer_clock2 = 0;
	timer_clock3 = 0;
}

//=============================================================================

