
#include "neopop.h"
#include "state.h"
#include "TLCS900H.h"
#include "interrupt.h"
#include "dma.h"
#include "mem.h"

//=============================================================================

static void read_state_0050(char* filename);

//-----------------------------------------------------------------------------
// state_restore()
//-----------------------------------------------------------------------------
void state_restore(char* filename)
{
	_u16 version;

	if (system_io_state_read(filename, (_u8*)&version, sizeof(_u16)))
	{
		switch(version)
		{
		case 0x0050:	read_state_0050(filename);	break;

		default:
			system_message("BAD STATE FILE");
			return;
		}

	}
}

//=============================================================================

//-----------------------------------------------------------------------------
// state_store()
//-----------------------------------------------------------------------------

extern tw tlcs;

void state_store(char* filename)
{
	NEOPOPSTATE0050	*state;
	int i,j;
	tw *rg=&tlcs;
	
	state = (NEOPOPSTATE0050	*)ljz_malloc(sizeof(NEOPOPSTATE0050));


	//Build a state description
	state->valid_state_id = 0x0050;
	memcpy(&(state->header), rom_header, sizeof(RomHeader));

	state->eepromStatusEnable = eepromStatusEnable;

	//TLCS-900h Registers
	state->pc = get_pc(rg);
	state->sr = tlcs.status;
	state->f_dash = f_dash;


	state->gpr[0]=regL4;
	state->gpr[1]=regL5;
	state->gpr[2]=regL6;
	state->gpr[3]=regL7;
	
	for (i = 0; i < 4; i++)
	{
		//state->gpr[i] = gpr[i];
		for (j = 0; j < 4; j++)
			state->gprBank[i][j] = tlcs.gprBank[i][j];
	}

	//Z80 Registers
	memcpy(&(state->Z80_regs), &Z80_regs, sizeof(Z80));

	//Sound Chips
	memcpy(&(state->toneChip), &toneChip, sizeof(SoundChip));
	memcpy(&(state->noiseChip), &noiseChip, sizeof(SoundChip));

	//Memory
	memcpy(&(state->ram[0]), ram, 0xC000);

	//Timers
	state->timer_hint = timer_hint;

	for (i = 0; i < 4; i++)	//Up-counters
		state->timer[i] = timer[i];

	state->timer_clock0 = timer_clock0;
	state->timer_clock1 = timer_clock1;
	state->timer_clock2 = timer_clock2;
	state->timer_clock3 = timer_clock3;

	//DMA
	for (i = 0; i < 4; i++)
	{
		state->dmaS[i] = dmaS[i];
		state->dmaD[i] = dmaD[i];
		state->dmaC[i] = dmaC[i];
		state->dmaM[i] = dmaM[i];
	}


	system_io_state_write(filename, (_u8*)state, sizeof(NEOPOPSTATE0050));
	
	ljz_free(state);
}

//=============================================================================

static void read_state_0050(char* filename)
{
	NEOPOPSTATE0050	*state;
	int i,j;
	tw *rg=&tlcs;
	
	state = (NEOPOPSTATE0050	*)ljz_malloc(sizeof(NEOPOPSTATE0050));

	if (system_io_state_read(filename, (_u8*)state, sizeof(NEOPOPSTATE0050)))
	{
		//Verify correct rom...
		if (memcmp(rom_header, &(state->header), sizeof(RomHeader)) != 0)
		{
			system_message("WRONG ROM");
			return;
		}

		//Apply state description
		reset();

		eepromStatusEnable = state->eepromStatusEnable;

		//TLCS-900h Registers
		set_pc(&tlcs,state->pc);
		tlcs.status = state->sr;				changedSP(&tlcs);
		f_dash = state->f_dash;

		eepromStatusEnable = state->eepromStatusEnable;


		regL4=state->gpr[0];
		regL5=state->gpr[1];
		regL6=state->gpr[2];
		regL7=state->gpr[3];
		for (i = 0; i < 4; i++)
		{
			//gpr[i] = state->gpr[i];
			for (j = 0; j < 4; j++)
				tlcs.gprBank[i][j] = state->gprBank[i][j];
		}

		//Timers
		timer_hint = state->timer_hint;

		for (i = 0; i < 4; i++)	//Up-counters
			timer[i] = state->timer[i];

		timer_clock0 = state->timer_clock0;
		timer_clock1 = state->timer_clock1;
		timer_clock2 = state->timer_clock2;
		timer_clock3 = state->timer_clock3;

		//Z80 Registers
		memcpy(&Z80_regs, &(state->Z80_regs), sizeof(Z80));

		//Sound Chips
		memcpy(&toneChip, &(state->toneChip), sizeof(SoundChip));
		memcpy(&noiseChip, &(state->noiseChip), sizeof(SoundChip));

		//DMA
		for (i = 0; i < 4; i++)
		{
			dmaS[i] = state->dmaS[i];
			dmaD[i] = state->dmaD[i];
			dmaC[i] = state->dmaC[i];
			dmaM[i] = state->dmaM[i];
		}

		//Memory
		memcpy(ram, &(state->ram[0]), 0xC000);
	}
	
	ljz_free(state);
}

//=============================================================================
