

#include "neopop.h"
#include "flash.h"
#include "interrupt.h"

//=============================================================================

RomInfo rom;
RomHeader* rom_header;

//=============================================================================

static void rom_hack(void)
{
	gfx_hack = FALSE;

	//=============================
	// GRAPHICS HACKS
	//=============================
	if ((rom_header->catalog == 89 && rom_header->subCatalog == 5)||	//Sonic
		(rom_header->catalog == 149 && rom_header->subCatalog == 36)||	//LastBlade(US)
		(rom_header->catalog == 100 && rom_header->subCatalog == 31)||	//LastBlade(J)
		(rom_header->catalog == 2 && rom_header->subCatalog == 8)||		//NG.Cup
		(rom_header->catalog == 57 && rom_header->subCatalog == 14)||	//NG.Cup+Colour
		(rom_header->catalog == 133 && rom_header->subCatalog == 8)||	//Ogre
		(rom_header->catalog == 148 && rom_header->subCatalog == 4)||	//Rockman
		(rom_header->catalog == 105 && rom_header->subCatalog == 21)||	//SNKvsCAPCOM
		(rom_header->catalog == 48 && rom_header->subCatalog == 26)||	//Sam.Shodown2
		(rom_header->catalog == 102 && rom_header->subCatalog == 241)||	//Wrestle-Beta
		(rom_header->catalog == 102 && rom_header->subCatalog == 2)||	//Pro-Wrestle
		(rom_header->catalog == 1 && rom_header->subCatalog == 10)||	//KOF-R1
		(rom_header->catalog == 35 && rom_header->subCatalog == 21))	//KOF-R2
	{
		gfx_hack = TRUE;
	}

	//=============================
	// SPECIFIC ROM HACKS !
	//=============================

	//"Neo-Neo! V1.0 (PD)"
	if (rom_header->catalog == 0 && rom_header->subCatalog == 16)
	{
		rom.data[0x23] = 0x10;	// Fix rom header
	}

	//"Cool Cool Jam SAMPLE (U)"
	if (rom_header->catalog == 4660 && rom_header->subCatalog == 161)
	{
		rom.data[0x23] = 0x10;	// Fix rom header
	}

	//"Dokodemo Mahjong (J)"
	if (rom_header->catalog == 51 && rom_header->subCatalog == 33)
	{
		rom.data[0x23] = 0x00;	// Fix rom header
	}

	//"Puyo Pop (V05) (JUE)"
	if (rom_header->catalog == 65 && rom_header->subCatalog == 5)
	{
		int i;
		for (i = 0x8F0; i < 0x8FC; i++)
			rom.data[i] = 0;
	}

	//"Puyo Pop (V06) (JUE)"
	if (rom_header->catalog == 65 && rom_header->subCatalog == 6)
	{
		int i;
		for (i = 0x8F0; i < 0x8FC; i++)
			rom.data[i] = 0;
	}

	//"Metal Slug - 2nd Mission (JUE) [!]"
	//"Metal Slug - 2nd Mission (JUE) [h1]"
	if (rom_header->catalog == 97 && rom_header->subCatalog == 4)
	{
		//Enable dev-kit code path, because otherwise it doesn't
		//allow jumping or firing (for some reason!)	
		rom.data[0x1f] = 0xFF;
		//Enables in-game voices ("Pineapple", etc.)
		//that were aren't supposed to be available in Dev-kit mode.
		rom.data[0x8DDF8] = 0xF0;	//28DDF7: "RET NZ" -> "RET F"
	}
}

//=============================================================================
/*
static void rom_display_header(void)
{
#ifdef NEOPOP_DEBUG
	system_debug_message("\nRom Header =========================");
	system_debug_message("\"%s\"", rom.name);

	//Colour / BW mode?
	switch(rom_header->mode)
	{
	case 0x00: system_debug_message("Presented in Black & White"); break;
	case 0x10: system_debug_message("Presented in Colour"); break;
	default: system_debug_message("Presented in Unknown mode %x", rom_header->mode); break;
	}

	//Catalog Numbers
	system_debug_message("Catalogue %d (sub %d)", rom_header->catalog, rom_header->subCatalog);

	//Starting PC
	system_debug_message("Starting PC = %06X", rom_header->startPC & 0xFFFFFF);

	system_debug_message("====================================\n");
#endif
}*/

//=============================================================================

//-----------------------------------------------------------------------------
// rom_loaded()
//-----------------------------------------------------------------------------
void rom_loaded(void)
{
	int i;

	//Extract the header
	rom_header = (RomHeader*)(rom.data);

	//Rom Name
	for(i = 0; i < 12; i++)
	{
		if (rom_header->name[i] >= 32 && rom_header->name[i] < 128)
			rom.name[i] = rom_header->name[i];
		else
			rom.name[i] = ' ';
	}
	rom.name[i] = 0;

	rom_hack();	//Apply a hack if reuqired!

//	rom_display_header();	//Show the header (debugger only)

	flash_read();
}

//-----------------------------------------------------------------------------
// rom_unload()
//-----------------------------------------------------------------------------
void rom_unload(void)
{
	if (rom.data)
	{
		int i;

		flash_commit();

		ljz_free(rom.data);
		rom.data = NULL;
		rom.length = 0;
		rom_header = 0;

		for (i = 0; i < 16; i++)
			rom.name[i] = 0;

		for (i = 0; i < 256; i++)
			rom.filename[i] = 0;

//		reset();

	}			
}

//=============================================================================
